#! /usr/bin/env python
"""
Build a lowest-common-ancestor database with given taxonomy and genome sigs.
"""

import sys
import csv
import os
from collections import defaultdict

from sourmash import sourmash_args
from sourmash.sourmash_args import load_file_as_signatures
from sourmash.logging import notify, error, debug, set_quiet
from . import lca_utils
from .lca_db import LCA_Database
from sourmash.sourmash_args import DEFAULT_LOAD_K


def load_taxonomy_assignments(
    filename,
    *,
    delimiter=",",
    start_column=2,
    use_headers=True,
    force=False,
    split_identifiers=False,
    keep_identifier_versions=False,
):
    """
    Load a taxonomy assignment spreadsheet into a dictionary.

    The 'assignments' dictionary that's returned maps identifiers to
    lineage tuples.
    """
    from sourmash.tax.tax_utils import LineagePair

    # parse spreadsheet!
    # CTB note: can't easily switch to FileInputCSV, because of
    # janky way we do/don't handle headers here. See issue #2198.
    fp = open(filename, newline="")
    r = csv.reader(fp, delimiter=delimiter)
    row_headers = ["identifiers"]
    row_headers += ["_skip_"] * (start_column - 2)
    row_headers += list(lca_utils.taxlist())

    # first check that headers are interpretable.
    if use_headers:
        notify("examining spreadsheet headers...")
        first_row = next(iter(r))

        n_disagree = 0
        for column, value in zip(row_headers, first_row):
            if column == "_skip_":
                continue

            if column.lower() != value.lower():
                notify(f"** assuming column '{value}' is {column} in spreadsheet")
                n_disagree += 1
                if n_disagree > 2:
                    error("whoa, too many assumptions. are the headers right?")
                    error("expecting {}", ",".join(row_headers))
                    if not force:
                        sys.exit(-1)
                    notify("...continue, because --force was specified.")

    # convert into a lineage pair
    assignments = {}
    num_rows = 0
    n_species = 0
    n_strains = 0
    for row in r:
        if row and row[0].strip():  # want non-empty row
            num_rows += 1
            lineage = list(zip(row_headers, row))
            lineage = [x for x in lineage if x[0] != "_skip_"]

            ident = lineage[0][1]
            lineage = lineage[1:]

            # fold, spindle, and mutilate ident?
            if split_identifiers:
                ident = ident.split(" ")[0]

                if not keep_identifier_versions:
                    ident = ident.split(".")[0]

            # clean lineage of null names, replace with 'unassigned'
            lineage = [(a, lca_utils.filter_null(b)) for (a, b) in lineage]
            lineage = [LineagePair(a, b) for (a, b) in lineage]

            # remove end nulls
            while lineage and lineage[-1].name == "unassigned":
                lineage = lineage[:-1]

            # store lineage tuple
            if lineage:
                # check duplicates
                if ident in assignments:
                    if assignments[ident] != tuple(lineage):
                        if not force:
                            raise Exception(f"multiple lineages for identifier {ident}")
                else:
                    assignments[ident] = tuple(lineage)

                    if lineage[-1].rank == "species":
                        n_species += 1
                    elif lineage[-1].rank == "strain":
                        n_species += 1
                        n_strains += 1

    fp.close()

    # this is to guard against a bug that happened once and I can't find
    # any more, when building a large GTDB-based database :) --CTB
    if len(assignments) * 0.2 > n_species and len(assignments) > 50:
        if not force:
            error("")
            error("ERROR: fewer than 20% of lineages have species-level resolution!?")
            error(
                "({} species assignments found, of {} assignments total)",
                n_species,
                len(assignments),
            )
            error("** If this is intentional, re-run the command with -f.")
            sys.exit(-1)

    return assignments, num_rows


def generate_report(
    record_duplicates,
    record_no_lineage,
    record_remnants,
    unused_lineages,
    unused_identifiers,
    filename,
):
    """
    Output a report of anomalies from building the index.
    """
    with open(filename, "w") as fp:
        print(f"Duplicate signatures: {len(record_duplicates)}", file=fp)
        fp.write("\n".join(record_duplicates))
        fp.write("\n")
        print(f"----\nUnused identifiers: {len(unused_identifiers)}", file=fp)
        fp.write("\n".join(unused_identifiers))
        fp.write("\n")
        print(
            f"----\nNo lineage provided for these identifiers: {len(record_no_lineage)}",
            file=fp,
        )
        fp.write("\n".join(record_no_lineage))
        fp.write("\n")
        print(
            f"----\nNo signatures found for these identifiers: {len(record_remnants)}",
            file=fp,
        )
        fp.write("\n".join(record_remnants))
        fp.write("\n")
        print(f"----\nUnused lineages: {len(unused_lineages)}", file=fp)
        for lineage in unused_lineages:
            fp.write(";".join(lca_utils.zip_lineage(lineage)))
            fp.write("\n")


def index(args):
    """
    main function for building an LCA database.
    """
    if args.start_column < 2:
        error("error, --start-column cannot be less than 2")
        sys.exit(-1)

    set_quiet(args.quiet, args.debug)

    args.scaled = int(args.scaled)

    if args.ksize is None:
        args.ksize = DEFAULT_LOAD_K

    moltype = sourmash_args.calculate_moltype(args, default="DNA")
    picklist = sourmash_args.load_picklist(args)

    db_outfile = args.lca_db_out
    if args.database_format == "json":
        if not (
            db_outfile.endswith(".lca.json") or db_outfile.endswith(".lca.json.gz")
        ):  # logic -> db.save
            db_outfile += ".lca.json"
    else:
        assert args.database_format == "sql"
        if not db_outfile.endswith(".lca.sql"):
            db_outfile += ".lca.sql"

    if os.path.exists(db_outfile):
        error(f"ERROR: output file {db_outfile} already exists. Not overwriting.")
        sys.exit(-1)

    notify(f"saving to LCA DB: {format(db_outfile)}")

    notify(
        f"Building LCA database with ksize={args.ksize} scaled={args.scaled} moltype={moltype}."
    )

    # first, load taxonomy spreadsheet
    delimiter = ","
    if args.tabs:
        delimiter = "\t"
    assignments, num_rows = load_taxonomy_assignments(
        args.csv,
        delimiter=delimiter,
        start_column=args.start_column,
        use_headers=not args.no_headers,
        force=args.force,
        split_identifiers=args.split_identifiers,
        keep_identifier_versions=args.keep_identifier_versions,
    )

    notify(
        f"{len(assignments)} distinct identities in spreadsheet out of {num_rows} rows."
    )
    notify(
        f"{len(set(assignments.values()))} distinct lineages in spreadsheet out of {num_rows} rows."
    )

    db = LCA_Database(args.ksize, args.scaled, moltype)

    inp_files = list(args.signatures)
    if args.from_file:
        more_files = sourmash_args.load_pathlist_from_file(args.from_file)
        inp_files.extend(more_files)

    # track duplicates
    md5_to_name = {}

    #
    # main loop, connecting lineage ID to signature.
    #

    n = 0
    total_n = len(inp_files)
    record_duplicates = set()
    record_no_lineage = []
    record_remnants = set(assignments)
    record_used_lineages = set()
    record_used_idents = set()
    n_skipped = 0
    for filename in inp_files:
        n += 1
        it = load_file_as_signatures(
            filename,
            ksize=args.ksize,
            select_moltype=moltype,
            picklist=picklist,
            yield_all_files=args.force,
        )
        for sig in it:
            notify("\r\033[K", end="")
            notify(
                f"\r... loading signature {str(sig)[:30]} ({n} of {total_n}); skipped {n_skipped} so far",
                end="",
            )
            debug(filename, sig)

            # block off duplicates.
            if sig.md5sum() in md5_to_name:
                debug(
                    "WARNING: in file {}, duplicate md5sum: {}; skipping",
                    filename,
                    sig.md5sum(),
                )
                record_duplicates.add(sig.name)
                continue

            md5_to_name[sig.md5sum()] = str(sig)

            # parse identifier, potentially with splitting
            if sig.name:
                ident = sig.name
            else:
                ident = sig.filename

            orig_ident = ident
            if args.split_identifiers:  # hack for NCBI-style names, etc.
                # split on space...
                ident = ident.split(" ")[0]

                if not args.keep_identifier_versions:
                    # ...and on period.
                    ident = ident.split(".")[0]

            lineage = assignments.get(ident)

            # punt if no lineage and --require-taxonomy
            if lineage is None and args.require_taxonomy:
                if args.fail_on_missing_taxonomy:
                    notify(f"ERROR: no taxonomy found for identifier '{ident}'")
                    if args.split_identifiers:
                        notify(f"(Identifier extracted from name: '{orig_ident})')")
                    sys.exit(-1)
                debug("(skipping, because --require-taxonomy was specified)")
                n_skipped += 1
                continue

            # add the signature into the database.
            try:
                db.insert(sig, ident=ident, lineage=lineage)
            except ValueError as e:
                error(
                    "ERROR: cannot insert signature '{}' (md5 {}, loaded from '{}') into database.",
                    sig,
                    sig.md5sum()[:8],
                    filename,
                )
                error("ERROR: {}", str(e))
                sys.exit(-1)

            if lineage:
                # remove from our list of remaining ident -> lineage
                record_remnants.remove(ident)

                # track ident as used
                record_used_idents.add(ident)
                record_used_lineages.add(lineage)

            # track lineage info - either no lineage, or this lineage used.
            else:
                debug("WARNING: no lineage assignment for {}.", ident)
                record_no_lineage.append(ident)

    # end main add signatures loop

    if n_skipped:
        notify(
            f"... loaded {total_n} signatures; skipped {n_skipped} because of --require-taxonomy."
        )
    else:
        notify(f"... loaded {total_n} signatures.")

    # check -- did we find any signatures?
    if n == 0:
        error("ERROR: no signatures found. ??")
        sys.exit(1)

    # check -- did the signatures we found have any hashes?
    if not db.hashvals:
        error("ERROR: no hash values found - are there any signatures?")
        sys.exit(1)
    notify(
        f"loaded {len(db.hashvals)} hashes at ksize={args.ksize} scaled={args.scaled}"
    )

    if picklist:
        sourmash_args.report_picklist(args, picklist)

    # summarize:
    notify(
        f"{len(record_used_lineages)} assigned lineages out of {len(set(assignments.values()))} distinct lineages in spreadsheet."
    )
    unused_lineages = set(assignments.values()) - record_used_lineages

    notify(
        f"{len(record_used_idents)} identifiers used out of {len(set(assignments))} distinct identifiers in spreadsheet."
    )

    assert record_used_idents.issubset(set(assignments))
    unused_identifiers = set(assignments) - record_used_idents

    # now, save!
    db.save(db_outfile, format=args.database_format)

    ## done!

    # output a record of stuff if requested/available:
    if record_duplicates or record_no_lineage or record_remnants or unused_lineages:
        if record_duplicates:
            notify(f"WARNING: {len(record_duplicates)} duplicate signatures.")
        if record_no_lineage:
            notify(
                f"WARNING: no lineage provided for {len(record_no_lineage)} signatures."
            )
        if record_remnants:
            notify(
                f"WARNING: no signatures for {len(record_remnants)} spreadsheet rows."
            )
        if unused_lineages:
            notify(f"WARNING: {len(unused_lineages)} unused lineages.")

        if unused_identifiers:
            notify(f"WARNING: {len(unused_identifiers)} unused identifiers.")

        if args.report:
            notify(f"generating a report and saving in '{args.report}'")
            generate_report(
                record_duplicates,
                record_no_lineage,
                record_remnants,
                unused_lineages,
                unused_identifiers,
                args.report,
            )
        else:
            notify("(You can use --report to generate a detailed report.)")


if __name__ == "__main__":
    sys.exit(index(sys.argv[1:]))



"""combine multiple Sequence Bloom Trees"""


def subparser(subparsers):
    subparser = subparsers.add_parser("sbt_combine")
    subparser.add_argument("sbt_name", help="name to save SBT into")
    subparser.add_argument("sbts", nargs="+", help="SBTs to combine to form a new SBT")
    subparser.add_argument("-x", "--bf-size", metavar="S", type=float, default=1e5)


def main(args):
    import sourmash

    return sourmash.commands.sbt_combine(args)



"""classify genomes"""


def subparser(subparsers):
    subparser = subparsers.add_parser("classify")
    subparser.add_argument(
        "--db", nargs="+", action="append", help="databases to use to classify"
    )
    subparser.add_argument(
        "--query",
        nargs="*",
        default=[],
        action="append",
        help="query signatures to classify",
    )
    subparser.add_argument(
        "--query-from-file", help="file containing list of signature files to query"
    )
    subparser.add_argument(
        "--threshold",
        metavar="T",
        type=int,
        default=5,
        help="minimum number of hashes needed for a taxonomic classification (default: 5)",
    )
    subparser.add_argument(
        "--majority",
        action="store_true",
        help="use majority vote classification instead of lca",
    )
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-d", "--debug", action="store_true", help="output debugging output"
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        default="-",
        help="output CSV to the specified file; by default output to stdout",
    )
    subparser.add_argument("--scaled", type=float)


def main(args):
    import sourmash

    return sourmash.lca.command_classify.classify(args)



import os

import pytest

import sourmash
from sourmash.signature import (
    SourmashSignature,
    FrozenSourmashSignature,
)
from sourmash.signature import (
    save_signatures_to_json,
    load_signatures_from_json,
    load_one_signature_from_json,
)
import sourmash_tst_utils as utils
from sourmash.minhash import MinHash, FrozenMinHash
from sourmash_tst_utils import SourmashCommandFailed


def test_minhash_copy(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    SourmashSignature(e, name="foo")
    f = e.copy()
    assert e == f


def test_sig_copy(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    sig1 = SourmashSignature(e, name="foo")
    sig2 = sig1.copy()
    assert sig1 == sig2


def test_sig_copy_frozen(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    sig1 = SourmashSignature(e, name="foo")
    sig2 = sig1.copy()
    assert sig1 == sig2
    with pytest.raises(TypeError) as e:
        sig2.minhash.add_hash(5)
    assert "FrozenMinHash does not support modification" in str(e.value)


def test_sig_copy_frozen_mutable(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    sig1 = SourmashSignature(e, name="foo")
    sig1.minhash = sig1.minhash.to_mutable()
    sig2 = sig1.copy()
    assert sig1 == sig2
    with pytest.raises(TypeError) as e:
        sig2.minhash.add_hash(5)
    assert "FrozenMinHash does not support modification" in str(e.value)


def test_compare(track_abundance):
    # same content, same name -> equal
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    SourmashSignature(e, name="foo")

    f = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    f.add_kmer("AT" * 10)
    SourmashSignature(f, name="foo")

    assert e == f


def test_compare_ne(track_abundance):
    # same content, different names -> different
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    sig1 = SourmashSignature(e, name="foo")

    f = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    f.add_kmer("AT" * 10)
    sig2 = SourmashSignature(f, name="bar")

    assert sig1 != sig2


def test_compare_ne2(track_abundance):
    # same content, different filename -> different
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    sig1 = SourmashSignature(e, name="foo", filename="a")

    f = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    f.add_kmer("AT" * 10)
    sig2 = SourmashSignature(f, name="foo", filename="b")

    assert sig1 != sig2
    assert sig2 != sig1


def test_compare_ne2_reverse(track_abundance):
    # same content, one has filename, other does not -> different
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    sig1 = SourmashSignature(e, name="foo")

    f = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    f.add_kmer("AT" * 10)
    sig2 = SourmashSignature(f, filename="b")

    assert sig2 != sig1
    assert sig1 != sig2


def test_hashable(track_abundance):
    # check: can we use signatures as keys in dictionaries and sets?
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)

    sig = SourmashSignature(e)

    x = set()
    x.add(sig)


def test_str(track_abundance):
    # signatures should be printable
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)

    sig = SourmashSignature(e)

    print(sig)
    assert repr(sig) == "SourmashSignature('', 59502a74)"

    sig._name = "fizbar"
    assert repr(sig) == "SourmashSignature('fizbar', 59502a74)"


def test_roundtrip(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    sig = SourmashSignature(e)
    s = save_signatures_to_json([sig])
    siglist = list(load_signatures_from_json(s))
    sig2 = siglist[0]

    assert sig.similarity(sig2) == 1.0
    assert sig2.similarity(sig) == 1.0
    assert isinstance(sig, SourmashSignature)
    assert not isinstance(sig, FrozenSourmashSignature)
    assert isinstance(sig2, FrozenSourmashSignature)

    assert isinstance(e, MinHash)
    assert isinstance(sig.minhash, FrozenMinHash)
    assert isinstance(sig2.minhash, FrozenMinHash)


def test_roundtrip_mutable_frozen(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    sig = SourmashSignature(e)
    assert isinstance(sig.minhash, FrozenMinHash)
    sig.minhash = sig.minhash.to_mutable()

    sig2 = sig.to_frozen()
    assert isinstance(sig2.minhash, FrozenMinHash)


def test_load_signature_ksize_nonint(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    sig = SourmashSignature(e)
    s = save_signatures_to_json([sig])
    siglist = list(load_signatures_from_json(s, ksize="20"))
    sig2 = siglist[0]

    assert sig.similarity(sig2) == 1.0
    assert sig2.similarity(sig) == 1.0


def test_roundtrip_empty(track_abundance):
    # edge case, but: empty minhash? :)
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)

    sig = SourmashSignature(e)
    s = save_signatures_to_json([sig])
    siglist = list(load_signatures_from_json(s))
    sig2 = siglist[0]

    assert sig.similarity(sig2) == 0
    assert sig2.similarity(sig) == 0


def test_roundtrip_scaled(track_abundance):
    e = MinHash(n=0, ksize=20, track_abundance=track_abundance, max_hash=2**61)
    e.add_hash(5)
    sig = SourmashSignature(e)
    s = save_signatures_to_json([sig])
    siglist = list(load_signatures_from_json(s))
    sig2 = siglist[0]
    e2 = sig2.minhash

    assert e.scaled == e2.scaled

    assert sig.similarity(sig2) == 1.0
    assert sig2.similarity(sig) == 1.0


def test_roundtrip_seed(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance, seed=10)
    e.add_hash(5)
    sig = SourmashSignature(e)
    s = save_signatures_to_json([sig])
    siglist = list(load_signatures_from_json(s))
    sig2 = siglist[0]
    e2 = sig2.minhash

    assert e.seed == e2.seed

    assert sig.similarity(sig2) == 1.0
    assert sig2.similarity(sig) == 1.0


def test_similarity_downsample(track_abundance):
    e = MinHash(n=0, ksize=20, track_abundance=track_abundance, max_hash=2**63)
    f = MinHash(n=0, ksize=20, track_abundance=track_abundance, max_hash=2**61)

    e.add_hash(1)
    e.add_hash(2**62)
    assert len(e.hashes) == 2

    f.add_hash(1)
    f.add_hash(2**62)  # should be discarded due to max_hash
    assert len(f.hashes) == 1

    ee = SourmashSignature(e)
    ff = SourmashSignature(f)

    with pytest.raises(ValueError) as e:  # mismatch in max_hash
        ee.similarity(ff)

    assert "mismatch in scaled; comparison fail" in str(e.value)

    x = ee.similarity(ff, downsample=True)
    assert round(x, 1) == 1.0


def test_add_sequence_bad_dna(track_abundance):
    # test add_sequence behavior on bad DNA
    mh = MinHash(n=1, ksize=21)
    sig = SourmashSignature(mh)

    with pytest.raises(ValueError) as e:
        sig.add_sequence("N" * 21, force=False)

    assert "invalid DNA character in input k-mer: NNNNNNNNNNNNNNNNNNNNN" in str(e.value)


def test_md5(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_hash(5)
    sig = SourmashSignature(e)
    assert sig.md5sum() == "eae27d77ca20db309e056e3d2dcd7d69", sig.md5sum()


def test_str_1(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    sig = SourmashSignature(e, name="foo")
    assert str(sig) == "foo"


def test_str_2(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    sig = SourmashSignature(e, filename="foo.txt")
    assert str(sig) == "foo.txt"


def test_str_3(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    sig = SourmashSignature(e, name="foo", filename="foo.txt")
    assert str(sig) == "foo"


def test_name_4(track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    sig = SourmashSignature(e)
    assert str(sig) == sig.md5sum()[:8]


def test_save_load_multisig(track_abundance):
    e1 = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    sig1 = SourmashSignature(e1)

    e2 = MinHash(n=1, ksize=25, track_abundance=track_abundance)
    sig2 = SourmashSignature(e2)

    x = save_signatures_to_json([sig1, sig2])
    y = list(load_signatures_from_json(x))

    print(x)

    assert len(y) == 2
    assert sig1 in y  # order not guaranteed, note.
    assert sig2 in y
    assert sig1 != sig2


def test_load_one_fail_nosig(track_abundance):
    x = save_signatures_to_json([])
    print((x,))
    with pytest.raises(ValueError):
        load_one_signature_from_json(x)


def test_load_one_succeed(track_abundance):
    e1 = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    sig1 = SourmashSignature(e1)

    x = save_signatures_to_json([sig1])

    y = load_one_signature_from_json(x)
    assert sig1 == y


def test_load_one_fail_multisig(track_abundance):
    e1 = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    sig1 = SourmashSignature(e1)

    e2 = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    sig2 = SourmashSignature(e2)

    x = save_signatures_to_json([sig1, sig2])

    with pytest.raises(ValueError):
        load_one_signature_from_json(x)


def test_save_minified(track_abundance):
    e1 = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    sig1 = SourmashSignature(e1, name="foo")

    e2 = MinHash(n=1, ksize=25, track_abundance=track_abundance)
    sig2 = SourmashSignature(e2, name="bar baz")

    x = save_signatures_to_json([sig1, sig2])
    assert b"\n" not in x
    assert len(x.split(b"\n")) == 1

    y = list(load_signatures_from_json(x))
    assert len(y) == 2
    assert any(sig.name == "foo" for sig in y)
    assert any(sig.name == "bar baz" for sig in y)


def test_load_minified(track_abundance):
    sigfile = utils.get_test_data("genome-s10+s11.sig")
    sigs = load_signatures_from_json(sigfile)

    minified = save_signatures_to_json(sigs)
    with open(sigfile) as f:
        orig_file = f.read()
    assert len(minified) < len(orig_file)
    assert b"\n" not in minified


def test_load_compressed(track_abundance):
    e1 = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    sig1 = SourmashSignature(e1)

    x = save_signatures_to_json([sig1], compression=5)

    y = load_one_signature_from_json(x)
    assert sig1 == y

    sigfile = utils.get_test_data("genome-s10+s11.sig.gz")
    load_signatures_from_json(sigfile)


def test_binary_fp(tmpdir, track_abundance):
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)

    path = tmpdir.join("1.sig")
    with open(str(path), "wb") as fp:
        sig = SourmashSignature(e)
        save_signatures_to_json([sig], fp)


def test_load_signatures_from_json_no_file_do_raise(tmpdir):
    path = tmpdir.join("dne.sig")
    siglist = load_signatures_from_json(path, do_raise=True)
    with pytest.raises(Exception):
        list(siglist)


def test_load_signatures_from_json_no_file_do_not_raise(tmpdir):
    path = tmpdir.join("dne.sig")
    siglist = load_signatures_from_json(path)
    siglist = list(siglist)
    assert not siglist


def test_max_containment():
    mh1 = MinHash(0, 21, scaled=1, track_abundance=False)
    mh2 = MinHash(0, 21, scaled=1, track_abundance=False)
    mh1.add_many((1, 2, 3, 4))
    mh2.add_many((1, 5))

    ss1 = SourmashSignature(mh1)
    ss2 = SourmashSignature(mh2)

    assert ss1.contained_by(ss2) == 1 / 4
    assert ss2.contained_by(ss1) == 1 / 2
    assert ss1.max_containment(ss2) == 1 / 2
    assert ss2.max_containment(ss1) == 1 / 2


def test_max_containment_empty():
    mh1 = MinHash(0, 21, scaled=1, track_abundance=False)
    mh2 = MinHash(0, 21, scaled=1, track_abundance=False)

    mh1.add_many((1, 2, 3, 4))

    ss1 = SourmashSignature(mh1)
    ss2 = SourmashSignature(mh2)

    assert ss1.contained_by(ss2) == 0
    assert ss2.contained_by(ss1) == 0
    assert ss1.max_containment(ss2) == 0
    assert ss2.max_containment(ss1) == 0


def test_max_containment_equal():
    mh1 = MinHash(0, 21, scaled=1, track_abundance=False)
    mh2 = MinHash(0, 21, scaled=1, track_abundance=False)

    mh1.add_many((1, 2, 3, 4))
    mh2.add_many((1, 2, 3, 4))

    ss1 = SourmashSignature(mh1)
    ss2 = SourmashSignature(mh2)

    assert ss1.contained_by(ss2) == 1
    assert ss2.contained_by(ss1) == 1
    assert ss1.max_containment(ss2) == 1
    assert ss2.max_containment(ss1) == 1


def test_containment_ANI():
    f1 = utils.get_test_data("2.fa.sig")
    f2 = utils.get_test_data("2+63.fa.sig")
    ss1 = sourmash.load_one_signature_from_json(f1, ksize=31)
    ss2 = sourmash.load_one_signature_from_json(f2, ksize=31)

    s1_cont_s2 = ss1.containment_ani(ss2, estimate_ci=True)
    s2_cont_s1 = ss2.containment_ani(ss1, estimate_ci=True)
    print("\nss1 contained by ss2", s1_cont_s2)
    print("ss2 contained by ss1", s2_cont_s1)

    assert (round(s1_cont_s2.ani, 3), s1_cont_s2.ani_low, s1_cont_s2.ani_high) == (
        1.0,
        1.0,
        1.0,
    )
    assert (
        round(s2_cont_s1.ani, 3),
        round(s2_cont_s1.ani_low, 3),
        round(s2_cont_s1.ani_high, 3),
    ) == (0.966, 0.965, 0.967)

    s1_mc_s2 = ss1.max_containment_ani(ss2, estimate_ci=True)
    s2_mc_s1 = ss2.max_containment_ani(ss1, estimate_ci=True)
    print("mh1 max containment", s1_mc_s2)
    print("mh2 max containment", s2_mc_s1)
    s1_mc_s2.size_is_inaccurate = False
    s2_mc_s1.size_is_inaccurate = False
    assert s1_mc_s2 == s2_mc_s1
    assert (
        round(s1_mc_s2.ani, 3),
        round(s1_mc_s2.ani_low, 3),
        round(s1_mc_s2.ani_high, 3),
    ) == (1.0, 1.0, 1.0)


def test_containment_ANI_precalc_containment():
    f1 = utils.get_test_data("47+63.fa.sig")
    f2 = utils.get_test_data("2+63.fa.sig")
    ss1 = sourmash.load_one_signature_from_json(f1, ksize=31)
    ss2 = sourmash.load_one_signature_from_json(f2, ksize=31)
    # precalc containments and assert same results
    s1c = ss1.contained_by(ss2)
    s2c = ss2.contained_by(ss1)
    mc = max(s1c, s2c)

    assert ss1.containment_ani(ss2, estimate_ci=True) == ss1.containment_ani(
        ss2, containment=s1c, estimate_ci=True
    )
    assert ss2.containment_ani(ss1) == ss2.containment_ani(ss1, containment=s2c)
    assert ss1.max_containment_ani(ss2) == ss2.max_containment_ani(ss1)
    assert ss1.max_containment_ani(ss2) == ss1.max_containment_ani(
        ss2, max_containment=mc
    )
    assert ss1.max_containment_ani(ss2) == ss2.max_containment_ani(
        ss1, max_containment=mc
    )


def test_avg_containment():
    f1 = utils.get_test_data("47+63.fa.sig")
    f2 = utils.get_test_data("2+63.fa.sig")
    ss1 = sourmash.load_one_signature_from_json(f1, ksize=31)
    ss2 = sourmash.load_one_signature_from_json(f2, ksize=31)
    # check average_containment_ani
    ac_s1 = ss1.avg_containment(ss2)
    ac_s2 = ss2.avg_containment(ss1)
    assert (
        ac_s1
        == ac_s2
        == (ss1.contained_by(ss2) + ss2.contained_by(ss1)) / 2
        == 0.6619979467456603
    )


def test_avg_containment_ani():
    f1 = utils.get_test_data("47+63.fa.sig")
    f2 = utils.get_test_data("2+63.fa.sig")
    ss1 = sourmash.load_one_signature_from_json(f1, ksize=31)
    ss2 = sourmash.load_one_signature_from_json(f2, ksize=31)
    # check average_containment_ani
    ac_s1 = ss1.avg_containment_ani(ss2)
    ac_s2 = ss2.avg_containment_ani(ss1)
    assert (
        ac_s1
        == ac_s2
        == (ss1.containment_ani(ss2).ani + ss2.containment_ani(ss1).ani) / 2
    )


def test_containment_ANI_downsample():
    f2 = utils.get_test_data("2+63.fa.sig")
    f3 = utils.get_test_data("47+63.fa.sig")
    ss2 = sourmash.load_one_signature_from_json(f2, ksize=31)
    ss3 = sourmash.load_one_signature_from_json(f3, ksize=31)
    # check that downsampling works properly
    print(ss2.minhash.scaled)

    ss2 = ss2.to_mutable()
    ss2.minhash = ss2.minhash.downsample(scaled=2000)
    assert ss2.minhash.scaled != ss3.minhash.scaled
    ds_s3c = ss2.containment_ani(ss3, downsample=True)
    ds_s4c = ss3.containment_ani(ss2, downsample=True)
    mc_w_ds_1 = ss2.max_containment_ani(ss3, downsample=True)
    mc_w_ds_2 = ss3.max_containment_ani(ss2, downsample=True)

    with pytest.raises(ValueError) as e:
        ss2.containment_ani(ss3)
        assert "ValueError: mismatch in scaled; comparison fail" in e

    with pytest.raises(ValueError) as e:
        ss2.max_containment_ani(ss3)
        assert "ValueError: mismatch in scaled; comparison fail" in e

    ss3 = ss3.to_mutable()
    ss3.minhash = ss3.minhash.downsample(scaled=2000)
    assert ss2.minhash.scaled == ss3.minhash.scaled
    ds_s3c_manual = ss2.containment_ani(ss3)
    ds_s4c_manual = ss3.containment_ani(ss2)
    ds_mc_manual = ss2.max_containment_ani(ss3)
    assert ds_s3c == ds_s3c_manual
    assert ds_s4c == ds_s4c_manual
    assert mc_w_ds_1 == mc_w_ds_2 == ds_mc_manual


def test_jaccard_ANI():
    f1 = utils.get_test_data("2.fa.sig")
    f2 = utils.get_test_data("2+63.fa.sig")
    ss1 = sourmash.load_one_signature_from_json(f1, ksize=31)
    ss2 = sourmash.load_one_signature_from_json(f2)

    print("\nJACCARD_ANI", ss1.jaccard_ani(ss2))

    s1_jani_s2 = ss1.jaccard_ani(ss2)
    s2_jani_s1 = ss2.jaccard_ani(ss1)

    assert s1_jani_s2 == s2_jani_s1
    assert (
        s1_jani_s2.ani,
        s1_jani_s2.p_nothing_in_common,
        s1_jani_s2.jaccard_error,
    ) == (0.9783711630110239, 0.0, 3.891666770716877e-07)


def test_jaccard_ANI_untrustworthy():
    f1 = utils.get_test_data("2.fa.sig")
    f2 = utils.get_test_data("2+63.fa.sig")
    ss1 = sourmash.load_one_signature_from_json(f1, ksize=31)
    ss2 = sourmash.load_one_signature_from_json(f2)

    print("\nJACCARD_ANI", ss1.jaccard_ani(ss2))

    s1_jani_s2 = ss1.jaccard_ani(ss2, err_threshold=1e-7)

    # since size is inaccurate on 2.fa.sig, need to override to be able to get ani
    s1_jani_s2.size_is_inaccurate = False

    assert s1_jani_s2.ani is None
    assert s1_jani_s2.je_exceeds_threshold == True
    assert s1_jani_s2.je_threshold == 1e-7


def test_jaccard_ANI_precalc_jaccard():
    f1 = utils.get_test_data("47+63.fa.sig")
    f2 = utils.get_test_data("2+63.fa.sig")
    ss1 = sourmash.load_one_signature_from_json(f1, ksize=31)
    ss2 = sourmash.load_one_signature_from_json(f2)
    # precalc jaccard and assert same result
    jaccard = ss1.jaccard(ss2)
    print("\nJACCARD_ANI", ss1.jaccard_ani(ss2, jaccard=jaccard))

    assert (
        ss1.jaccard_ani(ss2)
        == ss1.jaccard_ani(ss2, jaccard=jaccard)
        == ss2.jaccard_ani(ss1, jaccard=jaccard)
    )
    wrong_jaccard = jaccard - 0.1
    assert ss1.jaccard_ani(ss2) != ss1.jaccard_ani(ss2, jaccard=wrong_jaccard)


def test_jaccard_ANI_downsample():
    f1 = utils.get_test_data("47+63.fa.sig")
    f2 = utils.get_test_data("2+63.fa.sig")
    ss1 = sourmash.load_one_signature_from_json(f1, ksize=31)
    ss2 = sourmash.load_one_signature_from_json(f2)

    print(ss1.minhash.scaled)
    ss1 = ss1.to_mutable()
    ss1.minhash = ss1.minhash.downsample(scaled=2000)
    assert ss1.minhash.scaled != ss2.minhash.scaled
    with pytest.raises(ValueError) as e:
        ss1.jaccard_ani(ss2)
        assert "ValueError: mismatch in scaled; comparison fail" in e

    ds_s1c = ss1.jaccard_ani(ss2, downsample=True)
    ds_s2c = ss2.jaccard_ani(ss1, downsample=True)

    ss2 = ss2.to_mutable()
    ss2.minhash = ss2.minhash.downsample(scaled=2000)
    assert ss1.minhash.scaled == ss2.minhash.scaled
    ds_j_manual = ss1.jaccard_ani(ss2)
    assert ds_s1c == ds_s2c == ds_j_manual


def test_frozen_signature_update_1(track_abundance):
    # setting .name should fail on a FrozenSourmashSignature
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    ss = SourmashSignature(e, name="foo").to_frozen()

    with pytest.raises(ValueError):
        ss.name = "foo2"


def test_frozen_signature_update_2(track_abundance):
    # setting .minhash should fail on a FrozenSourmashSignature
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    e2 = e.copy_and_clear()
    ss = SourmashSignature(e, name="foo").to_frozen()

    with pytest.raises(ValueError):
        ss.minhash = e2


def test_frozen_signature_update_3(track_abundance):
    # setting .minhash should succeed with update() context manager
    e = MinHash(n=1, ksize=20, track_abundance=track_abundance)
    e.add_kmer("AT" * 10)
    ss = SourmashSignature(e, name="foo").to_frozen()

    with ss.update() as ss2:
        ss2.name = "foo2"

    assert ss2.name == "foo2"



"""show details of signature"""

usage = """

### `sourmash signature describe` - display detailed information about signatures

Display signature details.

For example,

sourmash sig describe tests/test-data/47.fa.sig

will display:

signature filename: tests/test-data/47.fa.sig
signature: NC_009665.1 Shewanella baltica OS185, complete genome
source file: 47.fa
md5: 09a08691ce52952152f0e866a59f6261
k=31 molecule=DNA num=0 scaled=1000 seed=42 track_abundance=0
size: 5177
signature license: CC0

"""

from sourmash.cli.utils import (
    add_moltype_args,
    add_ksize_arg,
    add_picklist_args,
    add_pattern_args,
)


def subparser(subparsers):
    subparser = subparsers.add_parser("describe", description=__doc__, usage=usage)
    subparser.add_argument("signatures", nargs="*")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-d", "--debug", action="store_true", help="provide debugging output"
    )
    subparser.add_argument(
        "--csv", metavar="FILE", help="output information to a CSV file"
    )
    subparser.add_argument(
        "-f", "--force", action="store_true", help="try to load all files as signatures"
    )
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_picklist_args(subparser)
    add_pattern_args(subparser)


def main(args):
    import sourmash

    return sourmash.sig.__main__.describe(args)



"""borrow abundances from one signature => one or more other signatures"""

from sourmash.cli.utils import add_moltype_args, add_ksize_arg, add_picklist_args


def subparser(subparsers):
    subparser = subparsers.add_parser("inflate")
    subparser.add_argument("signature_from")
    subparser.add_argument("other_sigs", nargs="+")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        default="-",
        help="output signature to this file (default stdout)",
    )
    subparser.add_argument(
        "-f", "--force", action="store_true", help="try to load all files as signatures"
    )
    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_picklist_args(subparser)


def main(args):
    import sourmash

    return sourmash.sig.__main__.inflate(args)



"""plot distance matrix made by 'compare'"""


def subparser(subparsers):
    subparser = subparsers.add_parser("plot")
    subparser.add_argument("distances", help='output from "sourmash compare"')
    subparser.add_argument(
        "--pdf", action="store_true", help="output PDF; default is PNG"
    )
    subparser.add_argument(
        "--labels",
        action="store_true",
        default=None,
        help="show sample labels on dendrogram/matrix",
    )
    subparser.add_argument(
        "--no-labels",
        action="store_false",
        dest="labels",
        help="do not show sample labels",
    )
    subparser.add_argument(
        "--labeltext",
        help="filename containing list of labels (overrides signature names); implies --labels",
    )
    subparser.add_argument(
        "--indices",
        action="store_true",
        default=None,
        help="show sample indices but not labels; overridden by --labels",
    )
    subparser.add_argument(
        "--no-indices",
        action="store_false",
        dest="indices",
        help="do not show sample indices",
    )
    subparser.add_argument(
        "--vmin",
        default=0.0,
        type=float,
        help="lower limit of heatmap scale; default=%(default)f",
    )
    subparser.add_argument(
        "--vmax",
        default=1.0,
        type=float,
        help="upper limit of heatmap scale; default=%(default)f",
    )
    subparser.add_argument(
        "--subsample",
        type=int,
        metavar="N",
        help="randomly downsample to this many samples, max",
    )
    subparser.add_argument(
        "--subsample-seed",
        type=int,
        default=1,
        metavar="S",
        help="random seed for --subsample; default=1",
    )
    subparser.add_argument(
        "-f", "--force", action="store_true", help="forcibly plot non-distance matrices"
    )
    subparser.add_argument(
        "--output-dir", metavar="DIR", help="directory for output plots"
    )
    subparser.add_argument(
        "--csv",
        metavar="F",
        help="write clustered matrix and labels out in CSV format (with column"
        " headers) to this file",
    )
    subparser.add_argument(
        "--labels-from",
        "--labels-load",
        help="a CSV file containing label information to use on plot; implies --labels",
    )


def main(args):
    import sourmash

    return sourmash.commands.plot(args)



"""summarize taxonomy/lineage information"""

usage = """

    sourmash tax summarize <taxonomy_file> [ <more files> ... ]

The 'tax summarize' command reads in one or more taxonomy databases
or lineage files (produced by 'tax annotate'), combines them,
and produces a human readable summary.

Please see the 'tax summarize' documentation for more details:
  https://sourmash.readthedocs.io/en/latest/command-line.html#command-line.html#sourmash-tax-summarize-print-summary-information-for-lineage-spreadsheets-or-taxonomy-databases

"""

import sourmash
from sourmash.logging import notify, print_results, error


def subparser(subparsers):
    subparser = subparsers.add_parser("summarize", usage=usage)
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "taxonomy_files",
        metavar="FILE",
        nargs="+",
        action="extend",
        help="database lineages",
    )
    subparser.add_argument(
        "-o",
        "--output-lineage-information",
        help="output a CSV file containing individual lineage counts",
    )
    subparser.add_argument(
        "--keep-full-identifiers",
        action="store_true",
        help="do not split identifiers on whitespace",
    )
    subparser.add_argument(
        "--keep-identifier-versions",
        action="store_true",
        help="after splitting identifiers, do not remove accession versions",
    )
    subparser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="continue past errors in file and taxonomy loading",
    )
    subparser.add_argument(
        "--lins",
        "--lin-taxonomy",
        action="store_true",
        default=False,
        help="use LIN taxonomy in place of standard taxonomic ranks.",
    )
    subparser.add_argument(
        "--ictv",
        "--ictv-taxonomy",
        action="store_true",
        default=False,
        help="use ICTV taxonomy in place of standard taxonomic ranks.  Note that the taxonomy CSV must contain ICTV ranks.",
    )


def main(args):
    return sourmash.tax.__main__.summarize(args)



"""
Test the behavior of LCA databases. New LCA database classes should support
all of this functionality.
"""

import pytest
import sourmash_tst_utils as utils

import sourmash
from sourmash.tax.tax_utils import MultiLineageDB
from sourmash.lca.lca_db import LCA_Database, load_single_database


def build_inmem_lca_db(runtmp):
    # test in-memory LCA_Database
    sigfile1 = utils.get_test_data(
        "prot/protein/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig"
    )
    sigfile2 = utils.get_test_data(
        "prot/protein/GCA_001593935.1_ASM159393v1_protein.faa.gz.sig"
    )

    ss1 = sourmash.load_one_signature(sigfile1)
    ss2 = sourmash.load_one_signature(sigfile2)

    lineages_file = utils.get_test_data("prot/gtdb-subset-lineages.csv")
    lineages = MultiLineageDB.load([lineages_file])

    db = LCA_Database(ksize=19, scaled=100, moltype="protein")

    ident1 = ss1.name.split(" ")[0].split(".")[0]
    assert lineages[ident1]
    db.insert(ss1, ident=ident1, lineage=lineages[ident1])
    ident2 = ss2.name.split(" ")[0].split(".")[0]
    assert lineages[ident2]
    db.insert(ss2, ident=ident2, lineage=lineages[ident2])

    return db


def build_json_lca_db(runtmp):
    # test saved/loaded JSON database
    db = build_inmem_lca_db(runtmp)
    db_out = runtmp.output("protein.lca.json")

    db.save(db_out, format="json")

    x = load_single_database(db_out)
    db_load = x[0]

    return db_load


def build_sql_lca_db(runtmp):
    # test saved/loaded SQL database
    db = build_inmem_lca_db(runtmp)
    db_out = runtmp.output("protein.lca.json")

    db.save(db_out, format="sql")

    x = load_single_database(db_out)
    db_load = x[0]

    return db_load


@pytest.fixture(params=[build_inmem_lca_db, build_json_lca_db, build_sql_lca_db])
def lca_db_obj(request, runtmp):
    build_fn = request.param

    return build_fn(runtmp)


def test_get_lineage_assignments(lca_db_obj):
    # test get_lineage_assignments for a specific hash
    lineages = lca_db_obj.get_lineage_assignments(178936042868009693)

    assert len(lineages) == 1
    lineage = lineages[0]

    x = []
    for tup in lineage:
        if tup[0] != "strain" or tup[1]:  # ignore empty strain
            x.append((tup[0], tup[1]))

    assert x == [
        ("superkingdom", "d__Archaea"),
        ("phylum", "p__Crenarchaeota"),
        ("class", "c__Bathyarchaeia"),
        ("order", "o__B26-1"),
        ("family", "f__B26-1"),
        ("genus", "g__B26-1"),
        ("species", "s__B26-1 sp001593925"),
    ]


def test_hashvals(lca_db_obj):
    # test getting individual hashvals
    hashvals = set(lca_db_obj.hashvals)
    assert 178936042868009693 in hashvals


def test_get_identifiers_for_hashval(lca_db_obj):
    # test getting identifiers belonging to individual hashvals
    idents = lca_db_obj.get_identifiers_for_hashval(178936042868009693)
    idents = list(idents)
    assert len(idents) == 1

    ident = idents[0]
    assert ident == "GCA_001593925"


def test_get_identifiers_for_hashval_2(lca_db_obj):
    # test systematic hashval => identifiers
    all_idents = set()

    for hashval in lca_db_obj.hashvals:
        idents = lca_db_obj.get_identifiers_for_hashval(hashval)
        # idents = list(idents)
        all_idents.update(idents)

    all_idents = list(all_idents)
    print(all_idents)
    assert len(all_idents) == 2

    assert "GCA_001593925" in all_idents
    assert "GCA_001593935" in all_idents


def test_downsample_scaled(lca_db_obj):
    # check the downsample_scaled method
    assert lca_db_obj.scaled == 100
    lca_db_obj.downsample_scaled(500)
    assert lca_db_obj.scaled == 500


def test_downsample_scaled_fail(lca_db_obj):
    # check the downsample_scaled method - should fail if lower scaled.
    assert lca_db_obj.scaled == 100

    with pytest.raises(ValueError):
        lca_db_obj.downsample_scaled(50)



"""
Tests for the 'CollectionManifest' class and protocol. All subclasses
of BaseCollectionManifest should support this functionality.
"""

import pytest
import sourmash_tst_utils as utils

import sourmash
from sourmash.manifest import BaseCollectionManifest, CollectionManifest
from sourmash.index.sqlite_index import SqliteCollectionManifest


def build_simple_manifest(runtmp):
    # load and return the manifest from prot/all.zip
    filename = utils.get_test_data("prot/all.zip")
    idx = sourmash.load_file_as_index(filename)
    mf = idx.manifest
    assert len(mf) == 8
    return mf


def build_sqlite_manifest(runtmp):
    # return the manifest from prot/all.zip
    filename = utils.get_test_data("prot/all.zip")
    idx = sourmash.load_file_as_index(filename)
    mf = idx.manifest

    # build sqlite manifest from this 'un
    mfdb = runtmp.output("test.sqlmf")
    return SqliteCollectionManifest.load_from_manifest(mf, dbfile=mfdb)


def save_load_manifest(runtmp):
    # save/load the manifest from a CSV.
    mf = build_simple_manifest(runtmp)

    mf_csv = runtmp.output("mf.csv")
    mf.write_to_filename(mf_csv)

    load_mf = CollectionManifest.load_from_filename(mf_csv)
    return load_mf


@pytest.fixture(
    params=[build_simple_manifest, save_load_manifest, build_sqlite_manifest]
)
def manifest_obj(request, runtmp):
    build_fn = request.param

    return build_fn(runtmp)


###
### generic CollectionManifeset tests go here
###


def test_manifest_len(manifest_obj):
    # check that 'len' works
    assert len(manifest_obj) == 8


def test_manifest_rows(manifest_obj):
    # check that '.rows' property works
    rows = list(manifest_obj.rows)
    assert len(rows) == 8

    required_keys = set(BaseCollectionManifest.required_keys)
    for row in rows:
        kk = set(row.keys())
        assert required_keys.issubset(kk)


def test_manifest_bool(manifest_obj):
    # check that 'bool' works
    assert bool(manifest_obj)


def test_make_manifest_row(manifest_obj):
    # build a manifest row from a signature
    sig47 = utils.get_test_data("47.fa.sig")
    ss = sourmash.load_one_signature(sig47)

    row = manifest_obj.make_manifest_row(ss, "foo", include_signature=False)
    assert "signature" not in row
    assert row["internal_location"] == "foo"

    assert row["md5"] == ss.md5sum()
    assert row["md5short"] == ss.md5sum()[:8]
    assert row["ksize"] == 31
    assert row["moltype"] == "DNA"
    assert row["num"] == 0
    assert row["scaled"] == 1000
    assert row["n_hashes"] == len(ss.minhash)
    assert not row["with_abundance"]
    assert row["name"] == ss.name
    assert row["filename"] == ss.filename


def test_manifest_create_manifest(manifest_obj):
    # test the 'create_manifest' method
    sig47 = utils.get_test_data("47.fa.sig")
    ss = sourmash.load_one_signature(sig47)

    def yield_sigs():
        yield ss, "fiz"

    new_mf = manifest_obj.create_manifest(yield_sigs(), include_signature=False)
    assert len(new_mf) == 1
    new_row = list(new_mf.rows)[0]

    row = manifest_obj.make_manifest_row(ss, "fiz", include_signature=False)

    required_keys = BaseCollectionManifest.required_keys
    for k in required_keys:
        assert new_row[k] == row[k], k


def test_manifest_select_to_manifest(manifest_obj):
    # do some light testing of 'select_to_manifest'
    new_mf = manifest_obj.select_to_manifest(moltype="DNA")
    assert len(new_mf) == 2


def test_manifest_locations(manifest_obj):
    # check the 'locations' method
    locs = set(
        [
            "dayhoff/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig",
            "dayhoff/GCA_001593935.1_ASM159393v1_protein.faa.gz.sig",
            "hp/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig",
            "hp/GCA_001593935.1_ASM159393v1_protein.faa.gz.sig",
            "protein/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig",
            "protein/GCA_001593935.1_ASM159393v1_protein.faa.gz.sig",
            "dna-sig.noext",
            "dna-sig.sig.gz",
        ]
    )
    assert set(manifest_obj.locations()) == locs


def test_manifest___contains__(manifest_obj):
    # check the 'in' operator
    sigfile = utils.get_test_data(
        "prot/dayhoff/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig"
    )
    ss = sourmash.load_one_signature(sigfile)

    assert ss in manifest_obj

    sigfile2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sigfile2, ksize=31)
    assert ss2 not in manifest_obj


def test_manifest_to_picklist(manifest_obj):
    # test 'to_picklist'
    picklist = manifest_obj.to_picklist()
    mf = manifest_obj.select_to_manifest(picklist=picklist)

    assert mf == manifest_obj


def test_manifest_filter_rows(manifest_obj):
    # test filter_rows
    def filter_fn(x):
        return "OS223" in x["name"]

    mf = manifest_obj.filter_rows(filter_fn)

    assert len(mf) == 1
    row = list(mf.rows)[0]
    assert row["name"] == "NC_011663.1 Shewanella baltica OS223, complete genome"


def test_manifest_filter_cols(manifest_obj):
    # test filter_rows
    def col_filter_fn(x):
        return "OS223" in x[0]

    mf = manifest_obj.filter_on_columns(col_filter_fn, ["name"])

    assert len(mf) == 1
    row = list(mf.rows)[0]
    assert row["name"] == "NC_011663.1 Shewanella baltica OS223, complete genome"


def test_manifest_iadd(manifest_obj):
    # test the 'create_manifest' method
    sig47 = utils.get_test_data("47.fa.sig")
    ss = sourmash.load_one_signature(sig47)

    def yield_sigs():
        yield ss, "fiz"

    new_mf = manifest_obj.create_manifest(yield_sigs(), include_signature=False)
    assert len(new_mf) == 1

    new_mf += manifest_obj
    assert len(new_mf) == len(manifest_obj) + 1


def test_manifest_add(manifest_obj):
    # test the 'create_manifest' method
    sig47 = utils.get_test_data("47.fa.sig")
    ss = sourmash.load_one_signature(sig47)

    def yield_sigs():
        yield ss, "fiz"

    new_mf = manifest_obj.create_manifest(yield_sigs(), include_signature=False)
    assert len(new_mf) == 1

    new_mf2 = new_mf + manifest_obj
    assert len(new_mf2) == len(manifest_obj) + len(new_mf)



"""create a manifest for a collection of signatures"""

usage = """

    sourmash sig manifest <filename> -o manifest.csv

This will output a sourmash manifest in CSV format. This manifest
can be used as a picklist with --picklist manifest.csv::manifest.

The manifest will be rebuilt by iterating over the signatures in the
file unless --no-rebuild-manifest is specified; for large
collections, rebuilding the manifest can take a long time!

See also the 'describe' and 'fileinfo' commands under 'sourmash sig'.

"""


def subparser(subparsers):
    subparser = subparsers.add_parser("manifest", usage=usage)
    subparser.add_argument("location")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-d", "--debug", action="store_true", help="output debug information"
    )
    subparser.add_argument(
        "-o",
        "--output",
        "--csv",
        metavar="FILE",
        help="output information to a CSV file",
        required=True,
    )
    subparser.add_argument(
        "-f", "--force", action="store_true", help="try to load all files as signatures"
    )
    subparser.add_argument(
        "--no-rebuild-manifest",
        help="use existing manifest if available",
        action="store_true",
    )
    subparser.add_argument(
        "-F",
        "--manifest-format",
        help="format of manifest output file; default is 'csv')",
        default="csv",
        choices=["csv", "sql"],
    )


def main(args):
    import sourmash

    return sourmash.sig.__main__.manifest(args)



"""
The main entry point for sourmash, defined in pyproject.toml.

Can also be executed as 'python -m sourmash'.
"""


def main(arglist=None):
    import sourmash

    args = sourmash.cli.parse_args(arglist)
    if hasattr(args, "subcmd"):
        mod = getattr(sourmash.cli, args.cmd)
        submod = getattr(mod, args.subcmd)
        mainmethod = getattr(submod, "main")
    else:
        mod = getattr(sourmash.cli, args.cmd)
        mainmethod = getattr(mod, "main")

    retval = mainmethod(args)
    raise SystemExit(retval)


if __name__ == "__main__":
    main()



"""sqlite3 based Index, CollectionManifest, and LCA_Database
implementations.

These classes support a variety of flexible and fast on-disk storage,
search, and retrieval functions.

SqliteIndex stores full scaled signatures; sketches are stored as
reverse-indexed collections of hashes. Search is optimized via the
reverse index. Num and abund sketches are not supported. All scaled
values must be the same upon insertion. Multiple moltypes _are_
supported.

SqliteCollectionManifest provides a full implementation of the
manifest API. It can store details for all signature types. When used
as part of a SqliteIndex database, it does not support independent
insertion.

LCA_SqliteDatabase builds on top of SqliteIndex and LineageDB_Sqlite
(in the tax submodule) to provide a full on-disk implementation of
LCA_Database.

Using these classes
-------------------

These classes are fully integrated into sourmash loading.

Internally, use `sqlite_index.load_sqlite_index(...)` to load a specific
file; this will return the appropriate SqliteIndex, StandaloneManifestIndex,
or LCA_Database object.

Use `CollectionManifest.load_from_filename(...)` to load the manifest
directly as a manifest object.

Implementation Details
----------------------

SqliteIndex:

* Hashes with values above MAX_SQLITE_INT=2**63-1 are transformed into
  signed long longs upon insertion, and then back into ulong longs upon
  retrieval.

* Hash overlap is calculated via a SELECT.

* SqliteIndex relies on SqliteCollectionManifest for manifest functionality,
  including signature selection and picklists.

SqliteCollectionManifest:

* each object maintains info about whether it is being "managed" by a
  SqliteIndex class or not. If it is, `_insert_row(...)` cannot be
  called directly.

* `select(...)` operates directly with SQL queries, except for
  picklist selection, which involves inspect each manifest row in
  Python. In addition to being (much) simpler, this ends up being
  faster in some important real world situations, even for millions of
  rows!

* filter_on_rows and filter_on_columns also both operate in Python,
  not SQL.

* for this reason, the `locations()` method returns a superset of
  locations.  This is potentially very significant if you do a select
  with a picklist that ignores most sketches - the `locations()`
  method will ignore the picklist.

Limitations:

* all of these classes share a single connection object, and it could
  get confusing quickly if you simultaneously insert and query. We suggest
  separating creation and insertion. That having been said, these databases
  should work fine for many simultaneous queries; just don't write :).

"""

import time
import os
import sqlite3
from collections import defaultdict
import itertools

from bitstring import BitArray

from sourmash.index import Index, _check_select_parameters
from sourmash.exceptions import IndexNotSupported
from sourmash import MinHash, SourmashSignature
from sourmash.index import IndexSearchResult, StandaloneManifestIndex
from sourmash.picklist import SignaturePicklist
from sourmash.logging import debug_literal
from sourmash import sqlite_utils

from sourmash.lca.lca_db import cached_property
from sourmash.manifest import BaseCollectionManifest

# converters for unsigned 64-bit ints: if over MAX_SQLITE_INT,
# convert to signed int.

MAX_SQLITE_INT = 2**63 - 1


def convert_hash_to(x):
    return BitArray(uint=x, length=64).int if x > MAX_SQLITE_INT else x


def convert_hash_from(x):
    return BitArray(int=x, length=64).uint if x < 0 else x


def load_sqlite_index(filename, *, request_manifest=False):
    """Load a SqliteIndex, SqliteCollectionManifest, or LCA_SqliteDatabase.

    This is the main top-level API for loading an Index-like object. The logic
    is roughly:

    * does this database have both index and lineage tables? If so,
      return an LCA_SqliteDatabase.
    * if it only has an index, return a SqliteIndex.
    * if it only has a manifest, return a StandaloneManifestIndex.

    If you would like only a manifest, specify 'request_manifest=True'.
    """
    conn = sqlite_utils.open_sqlite_db(filename)

    if conn is None:
        debug_literal("load_sqlite_index: conn is None.")
        return

    c = conn.cursor()
    internal_d = sqlite_utils.get_sourmash_internal(c)

    is_index = False
    is_manifest = False
    is_lca_db = False

    if "SqliteIndex" in internal_d:
        v = internal_d["SqliteIndex"]
        if v != "1.0":
            raise IndexNotSupported
        is_index = True
        debug_literal("load_sqlite_index: it's an index!")

    if is_index and "SqliteLineage" in internal_d:
        v = internal_d["SqliteLineage"]
        if v != "1.0":
            raise IndexNotSupported

        is_lca_db = True
        debug_literal("load_sqlite_index: it's got a lineage table!")

    if "SqliteManifest" in internal_d:
        v = internal_d["SqliteManifest"]
        if v != "1.0":
            raise IndexNotSupported
        is_manifest = True
        debug_literal(
            f"load_sqlite_index: it's a manifest! request_manifest: {request_manifest}"
        )

    # every Index is a Manifest!
    if is_index or is_lca_db:
        assert is_manifest

    idx = None
    if is_index and not request_manifest:
        conn.close()

        if is_lca_db:
            debug_literal("load_sqlite_index: returning LCA_SqliteDatabase")
            idx = LCA_SqliteDatabase.load(filename)
        else:
            debug_literal("load_sqlite_index: returning SqliteIndex")
            idx = SqliteIndex(filename)
    elif is_manifest:
        managed_by_index = False
        if is_index:
            assert request_manifest
            managed_by_index = True

        prefix = os.path.dirname(filename)
        mf = SqliteCollectionManifest(conn, managed_by_index=managed_by_index)
        idx = StandaloneManifestIndex(mf, filename, prefix=prefix)
        debug_literal("load_sqlite_index: returning StandaloneManifestIndex")

    return idx


class SqliteIndex(Index):
    is_database = True

    # NOTE: we do not need _signatures_with_internal for this class
    # because it supplies a manifest directly :tada:.

    def __init__(self, dbfile, *, sqlite_manifest=None, conn=None):
        "Constructor. 'dbfile' should be valid filename or ':memory:'."
        self.dbfile = dbfile

        # no connection? connect and/or create!
        if conn is None:
            conn = self._open(dbfile)

        # build me a SQLite manifest class to use for selection.
        if sqlite_manifest is None:
            sqlite_manifest = SqliteCollectionManifest(conn, managed_by_index=True)
        self.manifest = sqlite_manifest
        self.conn = conn

        # set 'scaled'.
        c = self.conn.cursor()
        c.execute("SELECT DISTINCT scaled FROM sourmash_sketches")
        scaled_vals = c.fetchall()
        if len(scaled_vals) > 1:
            raise ValueError(
                "this database has multiple scaled values, which is not currently allowed"
            )

        if scaled_vals:
            self.scaled = scaled_vals[0][0]
        else:
            self.scaled = None

    @classmethod
    def _open(cls, dbfile, *, empty_ok=True):
        "Connect to existing SQLite database or create new."
        try:
            conn = sqlite3.connect(dbfile)
            c = conn.cursor()

            c.execute("PRAGMA cache_size=10000000")
            c.execute("PRAGMA synchronous = OFF")
            c.execute("PRAGMA journal_mode = MEMORY")
            c.execute("PRAGMA temp_store = MEMORY")

            if not empty_ok:
                c.execute("SELECT * FROM sourmash_hashes LIMIT 1")
                c.fetchone()
        except (sqlite3.OperationalError, sqlite3.DatabaseError):
            raise ValueError(f"cannot open '{dbfile}' as SqliteIndex database")

        return conn

    @classmethod
    def load(self, dbfile):
        "Load an existing SqliteIndex from dbfile."
        return SqliteIndex(dbfile)

    @classmethod
    def create(cls, dbfile, *, append=False):
        "Create a new SqliteIndex in dbfile."
        conn = cls._open(dbfile, empty_ok=True)
        cls._create_tables(conn.cursor(), ignore_exists=append)
        conn.commit()

        return cls(dbfile, conn=conn)

    @classmethod
    def _create_tables(cls, c, *, ignore_exists=False):
        "Create sqlite tables for SqliteIndex"
        try:
            sqlite_utils.add_sourmash_internal(c, "SqliteIndex", "1.0")
            SqliteCollectionManifest._create_tables(c)

            c.execute(
                """
            CREATE TABLE IF NOT EXISTS sourmash_hashes (
               hashval INTEGER NOT NULL,
               sketch_id INTEGER NOT NULL,
               FOREIGN KEY (sketch_id) REFERENCES sourmash_sketches (id)
            )
            """
            )
            c.execute(
                """
            CREATE INDEX IF NOT EXISTS sourmash_hashval_idx ON sourmash_hashes (
               hashval,
               sketch_id
            )
            """
            )
            c.execute(
                """
            CREATE INDEX IF NOT EXISTS sourmash_hashval_idx2 ON sourmash_hashes (
               hashval
            )
            """
            )
            c.execute(
                """
            CREATE INDEX IF NOT EXISTS sourmash_sketch_idx ON sourmash_hashes (
               sketch_id
            )
            """
            )
        except (sqlite3.OperationalError, sqlite3.DatabaseError):
            if not ignore_exists:
                raise ValueError("cannot create SqliteIndex tables")

        return c

    def cursor(self):
        return self.conn.cursor()

    def close(self):
        self.conn.close()

    def commit(self):
        self.conn.commit()

    def __len__(self):
        return len(self.manifest)

    def insert(self, ss, *, cursor=None, commit=True):
        """
        Insert a signature into the sqlite database.

        If a cursor object is supplied, use that cursor instead of
        generating a new one.

        If 'commit' is True, commit after add; otherwise, do not.
        """
        if cursor:
            c = cursor
        else:
            c = self.conn.cursor()

        if ss.minhash.num:
            raise ValueError("cannot store 'num' signatures in SqliteIndex")
        if ss.minhash.track_abundance:
            raise ValueError("cannot store signatures with abundance in SqliteIndex")

        if self.scaled is not None and self.scaled != ss.minhash.scaled:
            raise ValueError(
                f"this database can only store scaled values={self.scaled}"
            )
        elif self.scaled is None:
            self.scaled = ss.minhash.scaled

        # ok, first create and insert a manifest row
        row = BaseCollectionManifest.make_manifest_row(
            ss, None, include_signature=False
        )
        self.manifest._insert_row(c, row, call_is_from_index=True)

        # retrieve ID of row for retrieving hashes:
        c.execute("SELECT last_insert_rowid()")
        (sketch_id,) = c.fetchone()

        # insert all the hashes
        hashes_to_sketch = []
        for h in ss.minhash.hashes:
            hh = convert_hash_to(h)
            hashes_to_sketch.append((hh, sketch_id))

        c.executemany(
            "INSERT INTO sourmash_hashes (hashval, sketch_id) VALUES (?, ?)",
            hashes_to_sketch,
        )

        if commit:
            self.conn.commit()

    @property
    def location(self):
        return self.dbfile

    def signatures(self):
        "Return an iterator over all signatures in the Index object."
        for ss, loc in self.signatures_with_location():
            yield ss

    def signatures_with_location(self):
        "Return an iterator over tuples (signature, location) in the Index."
        c = self.conn.cursor()

        for ss, loc, iloc in self._load_sketches(c):
            yield ss, loc

    def save(self, *args, **kwargs):
        raise NotImplementedError

    def find(self, search_fn, query, **kwargs):
        search_fn.check_is_compatible(query)

        # check compatibility, etc.
        query_mh = query.minhash
        if self.scaled > query_mh.scaled:
            query_mh = query_mh.downsample(scaled=self.scaled)

        picklist = None
        if self.manifest.selection_dict:
            picklist = self.manifest.selection_dict.get("picklist")

        c1 = self.conn.cursor()
        c2 = self.conn.cursor()

        debug_literal("running _get_matching_sketches...")
        t0 = time.time()
        xx = self._get_matching_sketches(c1, query_mh.hashes, query_mh._max_hash)
        for sketch_id, n_matching_hashes in xx:
            debug_literal(
                f"...got sketch {sketch_id}, with {n_matching_hashes} matching hashes in {time.time() - t0:.2f}"
            )
            #
            # first, estimate sketch size using sql results.
            #
            query_size = len(query_mh)
            subj_size = self._load_sketch_size(c2, sketch_id, query_mh._max_hash)
            total_size = query_size + subj_size - n_matching_hashes
            shared_size = n_matching_hashes

            score = search_fn.score_fn(query_size, shared_size, subj_size, total_size)

            debug_literal(
                f"APPROX RESULT: score={score} qsize={query_size}, ssize={subj_size} total={total_size} overlap={shared_size}"
            )

            # do we pass?
            if not search_fn.passes(score):
                debug_literal(f"FAIL score={score}")

            # CTB if we are doing containment only, we could break loop here.
            # but for Jaccard, we must continue.
            # see 'test_sqlite_jaccard_ordering'

            if search_fn.passes(score):
                subj = self._load_sketch(c2, sketch_id)
                if search_fn.collect(score, subj):
                    if picklist is None or subj in picklist:
                        yield IndexSearchResult(score, subj, self.location)

    def _select(self, *, num=0, track_abundance=False, **kwargs):
        _check_select_parameters(**kwargs)

        "Run a select! This just modifies the manifest."
        # check SqliteIndex specific conditions on the 'select'
        if num:
            raise ValueError("cannot select on 'num' in SqliteIndex")
        if track_abundance:
            raise ValueError("cannot store or search signatures with abundance")
        # create manifest if needed
        manifest = self.manifest
        if manifest is None:
            manifest = SqliteCollectionManifest(self.conn, managed_by_index=True)

        # modify manifest
        manifest = manifest.select_to_manifest(**kwargs)

        return manifest

    def select(self, *args, **kwargs):
        sqlite_manifest = self._select(*args, **kwargs)

        # return a new SqliteIndex with a new manifest, but same old conn.
        return SqliteIndex(self.dbfile, sqlite_manifest=sqlite_manifest, conn=self.conn)

    #
    # Actual SQL queries, etc.
    #

    def _load_sketch_size(self, c1, sketch_id, max_hash):
        "Get sketch size for given sketch, downsampled by max_hash."
        if max_hash <= MAX_SQLITE_INT:
            c1.execute(
                """
            SELECT COUNT(hashval) FROM sourmash_hashes
            WHERE sketch_id=? AND hashval >= 0 AND hashval <= ?""",
                (sketch_id, max_hash),
            )
        else:
            c1.execute(
                "SELECT COUNT(hashval) FROM sourmash_hashes WHERE sketch_id=?",
                (sketch_id,),
            )

        (n_hashes,) = c1.fetchone()
        return n_hashes

    def _load_sketch(self, c, sketch_id, *, match_scaled=None):
        "Load an individual sketch. If match_scaled is set, downsample."

        start = time.time()
        c.execute(
            """
        SELECT id, name, scaled, ksize, filename, moltype, seed
        FROM sourmash_sketches WHERE id=?""",
            (sketch_id,),
        )
        debug_literal(
            f"load sketch {sketch_id}: got sketch info in {time.time() - start:.2f}"
        )

        sketch_id, name, scaled, ksize, filename, moltype, seed = c.fetchone()
        if match_scaled is not None:
            scaled = max(scaled, match_scaled)

        is_protein = 1 if moltype == "protein" else 0
        is_dayhoff = 1 if moltype == "dayhoff" else 0
        is_hp = 1 if moltype == "hp" else 0

        mh = MinHash(
            n=0,
            ksize=ksize,
            scaled=scaled,
            seed=seed,
            is_protein=is_protein,
            dayhoff=is_dayhoff,
            hp=is_hp,
        )

        template_values = [sketch_id]

        hash_constraint_str = ""
        max_hash = mh._max_hash
        if max_hash <= MAX_SQLITE_INT:
            hash_constraint_str = (
                "sourmash_hashes.hashval >= 0 AND sourmash_hashes.hashval <= ? AND"
            )
            template_values.insert(0, max_hash)
        else:
            debug_literal("NOT EMPLOYING hash_constraint_str")

        debug_literal(
            f"finding hashes for sketch {sketch_id} in {time.time() - start:.2f}"
        )
        c.execute(
            f"SELECT hashval FROM sourmash_hashes WHERE {hash_constraint_str} sourmash_hashes.sketch_id=?",
            template_values,
        )

        debug_literal(
            f"loading hashes for sketch {sketch_id} in {time.time() - start:.2f}"
        )
        for (hashval,) in c:
            hh = convert_hash_from(hashval)
            mh.add_hash(hh)

        debug_literal(f"done loading sketch {sketch_id} {time.time() - start:.2f})")

        return SourmashSignature(mh, name=name, filename=filename)

    def _load_sketches(self, c):
        "Load sketches based on manifest _id column."
        for row in self.manifest.rows:
            sketch_id = row["_id"]
            assert row["num"] == 0

            moltype = row["moltype"]
            is_protein = 1 if moltype == "protein" else 0
            is_dayhoff = 1 if moltype == "dayhoff" else 0
            is_hp = 1 if moltype == "hp" else 0

            ksize = row["ksize"]
            scaled = row["scaled"]
            seed = row["seed"]

            mh = MinHash(
                n=0,
                ksize=ksize,
                scaled=scaled,
                seed=seed,
                is_protein=is_protein,
                dayhoff=is_dayhoff,
                hp=is_hp,
            )

            c.execute(
                "SELECT hashval FROM sourmash_hashes WHERE sketch_id=?", (sketch_id,)
            )

            for (hashval,) in c:
                mh.add_hash(convert_hash_from(hashval))

            ss = SourmashSignature(mh, name=row["name"], filename=row["filename"])
            yield ss, self.dbfile, sketch_id

    def _get_matching_sketches(self, c, hashes, max_hash):
        """
        For hashvals in 'hashes', retrieve all matching sketches,
        together with the number of overlapping hashes for each sketch.

        CTB: we do not use sqlite manifest conditions on this select,
        because it slows things down in practice.
        """
        c.execute("DROP TABLE IF EXISTS sourmash_hash_query")
        c.execute(
            "CREATE TEMPORARY TABLE sourmash_hash_query (hashval INTEGER PRIMARY KEY)"
        )

        hashvals = [(convert_hash_to(h),) for h in hashes]
        c.executemany(
            "INSERT OR IGNORE INTO sourmash_hash_query (hashval) VALUES (?)", hashvals
        )

        #
        # set up SELECT conditions
        #

        conditions = []
        template_values = []

        # downsample? => add to conditions
        max_hash = min(max_hash, max(hashes))
        if max_hash <= MAX_SQLITE_INT:
            select_str = "sourmash_hashes.hashval >= 0 AND sourmash_hashes.hashval <= ?"
            conditions.append(select_str)
            template_values.append(max_hash)

        # format conditions
        conditions.append("sourmash_hashes.hashval=sourmash_hash_query.hashval")
        conditions = " AND ".join(conditions)

        c.execute(
            f"""
        SELECT DISTINCT sourmash_hashes.sketch_id,COUNT(sourmash_hashes.hashval) as CNT
        FROM sourmash_hashes, sourmash_hash_query
        WHERE {conditions}
        GROUP BY sourmash_hashes.sketch_id ORDER BY CNT DESC
        """,
            template_values,
        )

        return c


class SqliteCollectionManifest(BaseCollectionManifest):
    """
    A SQLite-based manifest, used both for SqliteIndex and as a standalone
    manifest class.

    This class serves two purposes:
    * first, it is a fast, on-disk manifest that can be used in place of
      CollectionManifest.
    * second, it can be included within a SqliteIndex (which stores hashes
      too). In this case, however, new entries must be inserted by SqliteIndex
      rather than directly in this class.

    In the latter case, the SqliteCollectionManifest is created with
    managed_by_index set to True.
    """

    def __init__(self, conn, *, selection_dict=None, managed_by_index=False):
        """
        Here, 'conn' should already be connected and configured.

        Use 'create(filename)' to create a new database.

        Use 'create_from_manifest(filename, manifest) to create a new db
        from an existing manifest object.

        Use 'load_from_filename' to load from file.
        """
        assert conn is not None
        self.conn = conn
        self.selection_dict = selection_dict
        self.managed_by_index = managed_by_index
        self._num_rows = None

    @classmethod
    def create(cls, filename):
        "Connect to 'filename' and create the tables as a standalone manifest."
        conn = sqlite3.connect(filename)
        cursor = conn.cursor()
        cls._create_tables(cursor)
        return cls(conn)

    @classmethod
    def create_or_open(cls, filename):
        "Connect to 'filename' and create tables if not exist."
        conn = sqlite3.connect(filename)
        cursor = conn.cursor()
        try:
            cls._create_tables(cursor)
        except sqlite3.OperationalError:
            pass
        return cls(conn)

    @classmethod
    def load_from_manifest(cls, manifest, *, dbfile=":memory:", append=False):
        "Create a new sqlite manifest from an existing manifest object."
        return cls._create_manifest_from_rows(
            manifest.rows, location=dbfile, append=append
        )

    @classmethod
    def create_manifest(cls, locations_iter, *, include_signature=False):
        """Create a manifest from an iterator that yields (ss, location)

        Stores signatures in manifest rows by default.

        Note: do NOT catch exceptions here, so this passes through load excs.
        Note: this method ignores 'include_signature'.
        """

        def rows_iter():
            for ss, location in locations_iter:
                row = cls.make_manifest_row(ss, location, include_signature=False)
                yield row

        return cls._create_manifest_from_rows(rows_iter())

    @classmethod
    def _create_tables(cls, cursor):
        "Create the manifest table."
        # this is a class method so that it can be used by SqliteIndex to
        # create manifest-compatible tables.

        sqlite_utils.add_sourmash_internal(cursor, "SqliteManifest", "1.0")
        cursor.execute(
            """
        CREATE TABLE sourmash_sketches
          (id INTEGER PRIMARY KEY,
           name TEXT,
           num INTEGER NOT NULL,
           scaled INTEGER NOT NULL,
           ksize INTEGER NOT NULL,
           filename TEXT,
           moltype TEXT NOT NULL,
           with_abundance BOOLEAN NOT NULL,
           md5sum TEXT NOT NULL,
           seed INTEGER NOT NULL,
           n_hashes INTEGER NOT NULL,
           internal_location TEXT,
        UNIQUE(internal_location, md5sum)
        )
        """
        )

    def add_row(self, row):
        c = self.conn.cursor()
        self._insert_row(c, row)

    def _insert_row(self, cursor, row, *, call_is_from_index=False):
        "Insert a new manifest row."
        # check - is this manifest managed by SqliteIndex? If so, prevent
        # insertions unless SqliteIndex is the one calling it.
        if self.managed_by_index and not call_is_from_index:
            raise Exception("must use SqliteIndex.insert to add to this manifest")

        row = dict(row)
        if "seed" not in row:
            row["seed"] = 42

        cursor.execute(
            """
        INSERT OR IGNORE INTO sourmash_sketches
          (name, num, scaled, ksize, filename, md5sum, moltype,
           seed, n_hashes, with_abundance, internal_location)
        VALUES (:name, :num, :scaled, :ksize, :filename, :md5,
                :moltype, :seed, :n_hashes, :with_abundance,
                :internal_location)""",
            row,
        )

        self._num_rows = None  # reset cache

    def __bool__(self):
        "Is this manifest empty?"
        if self._num_rows is not None:
            return bool(self._num_rows)

        try:
            next(iter(self.rows))
            return True
        except StopIteration:
            return False

    def __eq__(self, other):
        "Check equality on a row-by-row basis. May fail on out-of-order rows."
        for a, b in itertools.zip_longest(self.rows, other.rows):
            # ignore non-required keys.
            for k in self.required_keys:
                if a[k] != b[k]:
                    return False

        return True

    def __len__(self):
        "Number of rows."

        # can we use cached value?
        if self._num_rows is not None:
            return self._num_rows

        # self.rows is a generator, so can't use 'len'
        self._num_rows = sum(1 for _ in self.rows)
        return self._num_rows

    def __iadd__(self, other):
        c = self.conn.cursor()
        for row in other.rows:
            self._insert_row(c, row)
        return self

    def __add__(self, other):
        new_mf = self.create(":memory:")
        new_mf += self
        new_mf += other
        return new_mf

    def close(self):
        self.conn.commit()

    def _make_select(self):
        """Build a set of SQL SELECT conditions and matching value tuple
        that can be used to select the right sketches from the
        database.

        Returns a triple 'conditions', 'values', and 'picklist'.
        'conditions' is a list that should be joined with 'AND'.

        The picklist is simply retrieved from the selection dictionary.
        """
        conditions = []
        values = []
        picklist = None
        if self.selection_dict:
            select_d = self.selection_dict
            if "ksize" in select_d and select_d["ksize"]:
                conditions.append("sourmash_sketches.ksize = ?")
                values.append(select_d["ksize"])
            if "num" in select_d and select_d["num"] > 0:
                conditions.append("sourmash_sketches.num > 0")
            if "scaled" in select_d and select_d["scaled"] > 0:
                conditions.append("sourmash_sketches.scaled > 0")
            if "containment" in select_d and select_d["containment"]:
                conditions.append("sourmash_sketches.scaled > 0")
            if "moltype" in select_d and select_d["moltype"] is not None:
                moltype = select_d["moltype"]
                assert moltype in ("DNA", "protein", "dayhoff", "hp"), moltype
                conditions.append(f"sourmash_sketches.moltype = '{moltype}'")

            picklist = select_d.get("picklist")

        return conditions, values, picklist

    def select_to_manifest(self, **kwargs):
        "Create a new SqliteCollectionManifest with the given select args."
        # Pass along all the selection kwargs to a new instance
        if self.selection_dict:
            debug_literal("sqlite manifest: merging selection dicts")
            # combine selects...
            d = dict(self.selection_dict)
            for k, v in kwargs.items():
                if k in d:
                    if d[k] is not None and d[k] != v:
                        raise ValueError(f"incompatible select on '{k}'")
                d[k] = v
            kwargs = d

        new_mf = SqliteCollectionManifest(self.conn, selection_dict=kwargs)

        # if picklist, make sure we fill in 'found'.
        picklist = kwargs.get("picklist")
        if picklist is not None:
            debug_literal("sqlite manifest: iterating through picklist")
            _ = len(self)  # this forces iteration through rows.

        return new_mf

    @property
    def rows(self):
        "Return rows that match the selection."
        c1 = self.conn.cursor()

        conditions, values, picklist = self._make_select()
        if conditions:
            conditions = conditions = "WHERE " + " AND ".join(conditions)
        else:
            conditions = ""

        debug_literal(f"sqlite manifest rows: executing select with '{conditions}'")
        c1.execute(
            f"""
        SELECT id, name, md5sum, num, scaled, ksize, filename, moltype,
        seed, n_hashes, internal_location FROM sourmash_sketches {conditions}
        """,
            values,
        )

        debug_literal("sqlite manifest: entering row yield loop")
        for (
            _id,
            name,
            md5sum,
            num,
            scaled,
            ksize,
            filename,
            moltype,
            seed,
            n_hashes,
            iloc,
        ) in c1:
            row = dict(
                num=num,
                scaled=scaled,
                name=name,
                filename=filename,
                n_hashes=n_hashes,
                with_abundance=False,
                ksize=ksize,
                md5=md5sum,
                internal_location=iloc,
                moltype=moltype,
                md5short=md5sum[:8],
                seed=seed,
                _id=_id,
            )
            if picklist is None or picklist.matches_manifest_row(row):
                yield row

    def filter_rows(self, row_filter_fn):
        """Create a new manifest filtered through row_filter_fn.

        This is done in memory, inserting each row one at a time.
        """

        def rows_iter():
            for row in self.rows:
                if row_filter_fn(row):
                    yield row

        return self._create_manifest_from_rows(rows_iter())

    def filter_on_columns(self, col_filter_fn, col_names):
        "Create a new manifest based on column matches."

        def row_filter_fn(row):
            x = [row[col] for col in col_names if row[col] is not None]
            return col_filter_fn(x)

        return self.filter_rows(row_filter_fn)

    def locations(self):
        """Return all possible locations for signatures.

        CTB: this may be a (big) superset of locations, if picklists are used.
        See test_sqlite_manifest_locations.

        Use set(row['internal_locations'] for row in self.rows)
        if you want an exact set of locations; will be slow for big manifests
        tho.
        """
        c1 = self.conn.cursor()

        conditions, values, picklist = self._make_select()
        if conditions:
            conditions = conditions = "WHERE " + " AND ".join(conditions)
        else:
            conditions = ""

        c1.execute(
            f"""
        SELECT DISTINCT internal_location FROM sourmash_sketches {conditions}
        """,
            values,
        )

        return (iloc for (iloc,) in c1)

    def __contains__(self, ss):
        "Check to see if signature 'ss' is in this manifest."
        md5 = ss.md5sum()

        c = self.conn.cursor()
        c.execute("SELECT COUNT(*) FROM sourmash_sketches WHERE md5sum=?", (md5,))
        (val,) = c.fetchone()

        if bool(val):
            picklist = self.picklist
            return picklist is None or ss in self.picklist
        return False

    @property
    def picklist(self):
        "Return the picklist, if any."
        if self.selection_dict:
            return self.selection_dict.get("picklist")
        return None

    def to_picklist(self):
        "Convert this manifest to a picklist."
        pl = SignaturePicklist("manifest")
        pl.pickset = {pl._get_value_for_manifest_row(row) for row in self.rows}
        return pl

    @classmethod
    def _create_manifest_from_rows(
        cls, rows_iter, *, location=":memory:", append=False
    ):
        """Create a SqliteCollectionManifest from a rows iterator.

        Internal utility function.

        CTB: should enable converting in-memory sqlite db to on-disk,
        probably with sqlite3 'conn.backup(...)' function.
        """
        try:
            mf = cls.create(location)
        except (sqlite3.OperationalError, sqlite3.DatabaseError) as exc:
            if not append:
                raise Exception(
                    f"cannot create sqlite3 db at '{location}'; exception: {str(exc)}"
                )
            db = load_sqlite_index(location, request_manifest=True)
            mf = db.manifest

        cursor = mf.conn.cursor()

        for row in rows_iter:
            mf._insert_row(cursor, row)

        mf.conn.commit()
        return mf


class LCA_SqliteDatabase(SqliteIndex):
    """
    A wrapper class for SqliteIndex + lineage db => LCA_Database functionality.
    """

    is_database = True

    def __init__(self, dbfile, *, lineage_db=None, sqlite_manifest=None):
        # CTB note: we need to let SqliteIndex open dbfile here, so can't
        # just pass in a conn.
        super().__init__(dbfile, sqlite_manifest=sqlite_manifest)

        c = self.conn.cursor()

        c.execute("SELECT DISTINCT ksize, moltype FROM sourmash_sketches")
        res = list(c)
        if len(res) > 1:
            raise TypeError(
                "can only have one ksize & moltype in an LCA_SqliteDatabase"
            )
        if len(res) == 0:
            raise ValueError("cannot load an LCA_SqliteDatabase")

        self.ksize, self.moltype = res[0]
        debug_literal(f"setting ksize and moltype to {self.ksize}, {self.moltype}")

        if lineage_db is not None:
            self.lineage_db = lineage_db

            ## the below is done once, but could be implemented as something
            ## ~dynamic.
            self._build_index()

    @classmethod
    def load(cls, filename):
        "Load LCA_SqliteDatabase from a single file."
        from sourmash.tax.tax_utils import LineageDB_Sqlite

        # first, load the SqliteIndex:
        try:
            debug_literal("sqlite_index: loading LCA_SqliteDatabase as SqliteIndex.")
            obj = cls(filename)
        except sqlite3.OperationalError:
            raise ValueError(f"cannot open '{filename}' as a SQLite index.")

        # now, toss in the lineage DB.
        lineage_db = LineageDB_Sqlite(obj.conn)
        obj.lineage_db = lineage_db
        obj._build_index()

        return obj

    @classmethod
    def create(cls, filename, idx, lineage_db):
        "Create a LCA_SqliteDatabase in a single file from existing idx/ldb."
        from sourmash.tax.tax_utils import MultiLineageDB

        # first, save/create signatures...
        sqlidx = SqliteIndex.create(filename)

        for ss in idx.signatures():
            sqlidx.insert(ss)

        # now, save the lineage_db into the same database
        out_lineage_db = MultiLineageDB()
        out_lineage_db.add(lineage_db)
        out_lineage_db._save_sqlite(None, conn=sqlidx.conn)

        # and voila! return, I guess?
        return cls.load(filename)

    def _build_index(self):
        "Rebuild the mappings that support identifier <-> lineage."
        mf = self.manifest
        lineage_db = self.lineage_db

        ident_to_idx = {}
        next_lid = 0
        idx_to_lid = {}
        lineage_to_lid = {}
        lid_to_lineage = {}

        for row in mf.rows:
            name = row["name"]
            if name:
                # this is a bit of a hack. we try identifiers _with_ and
                # _without_ versions, and take whichever works. There is
                # definitely a better way to do this, but I can't think
                # of one right now.
                ident = name.split(" ")[0]

                lineage = lineage_db.get(ident)  # try with identifier version
                if lineage is None:  # nope - remove version.x
                    ident = name.split(".")[0]
                    lineage = lineage_db.get(ident)

                idx = row["_id"]  # this is only present in sqlite manifests.
                ident_to_idx[ident] = idx

                if lineage:
                    lid = lineage_to_lid.get(lineage)

                    # manufacture new lid?
                    if lid is None:
                        lid = next_lid
                        next_lid += 1

                        lineage_to_lid[lineage] = lid
                        lid_to_lineage[lid] = lineage

                    # assign idx <-> lid
                    idx_to_lid[idx] = lid

        self.ident_to_idx = ident_to_idx
        self.idx_to_lid = idx_to_lid
        self.lid_to_lineage = lid_to_lineage

    # prevent insertions
    def insert(self, *args, **kwargs):
        raise NotImplementedError

    # return correct type on select
    def select(self, *args, **kwargs):
        sqlite_manifest = self._select(*args, **kwargs)

        return LCA_SqliteDatabase(
            self.dbfile, sqlite_manifest=sqlite_manifest, lineage_db=self.lineage_db
        )

    ### LCA_Database API/protocol.

    def downsample_scaled(self, scaled):
        "Downsample the scaled for querying."
        if scaled < self.scaled:
            raise ValueError(f"cannot decrease scaled from {self.scaled} to {scaled}")

        # CTB: maybe return a new LCA_Database? Right now this isn't how
        # the lca_db protocol works tho.
        self.scaled = scaled

    def get_lineage_assignments(self, hashval, *, min_num=None):
        """
        Get a list of lineages for this hashval.
        """
        x = []

        idx_list = self.hashval_to_idx.get(hashval, [])
        if min_num is None or len(idx_list) >= min_num:
            for idx in idx_list:
                lid = self.idx_to_lid.get(idx, None)
                if lid is not None:
                    lineage = self.lid_to_lineage[lid]
                    x.append(lineage)

        return x

    @cached_property
    def idx_to_ident(self):
        "Map individual idx to ident."
        d = defaultdict(set)
        for ident, idx in self.ident_to_idx.items():
            assert idx not in d
            d[idx] = ident
        return d

    @property
    def hashval_to_idx(self):
        "Dynamically interpret the SQL 'hashes' table like it's a dict."
        return _SqliteIndexHashvalToIndex(self)

    @property
    def hashvals(self):
        "Return all hashvals"
        return iter(_SqliteIndexHashvalToIndex(self))

    def get_identifiers_for_hashval(self, hashval):
        "Return identifiers associated with this hashval"
        idxlist = self.hashval_to_idx[hashval]
        for idx in idxlist:
            yield self.idx_to_ident[idx]


class _SqliteIndexHashvalToIndex:
    """
    Internal wrapper class to retrieve keys and key/value pairs for
    hashval -> [ list of idx ].
    """

    def __init__(self, sqlidx):
        self.sqlidx = sqlidx

    def __iter__(self):
        "Get all hashvals."
        c = self.sqlidx.conn.cursor()
        c.execute("SELECT DISTINCT hashval FROM sourmash_hashes")
        for (hashval,) in c:
            yield hashval

    def get(self, key, dv=None):
        "Retrieve idxlist for a given hash."
        sqlidx = self.sqlidx
        c = sqlidx.cursor()

        hh = convert_hash_to(key)

        c.execute("SELECT sketch_id FROM sourmash_hashes WHERE hashval=?", (hh,))

        x = [convert_hash_from(h) for (h,) in c]
        return x or dv

    def __getitem__(self, key):
        "Retrieve idxlist for a given hash; raise KeyError if not present."
        v = self.get(key)
        if v is None:
            raise KeyError(key)
        return v



"""
Test the plugin framework in sourmash.plugins, which uses importlib.metadata
entrypoints.
"""

import sys
import pytest
import collections

import sourmash
from sourmash.logging import set_quiet

import sourmash_tst_utils as utils
from sourmash import plugins
from sourmash.index import LinearIndex
from sourmash.save_load import Base_SaveSignaturesToLocation, SaveSignaturesToLocation


_Dist = collections.namedtuple("_Dist", ["version"])


class FakeEntryPoint:
    """
    A class that stores a name and an object to be returned on 'load()'.
    Mocks the EntryPoint class used by importlib.metadata.
    """

    module = "test_plugin_framework"
    dist = _Dist("0.1")
    group = "groupfoo"

    def __init__(self, name, load_obj, *, error_on_import=None):
        self.name = name
        self.load_obj = load_obj
        self.error_on_import = error_on_import

    def load(self):
        if self.error_on_import is not None:
            raise self.error_on_import("as requested")
        return self.load_obj


#
# Test basic features of the load_from plugin hook.
#


class Test_EntryPointBasics_LoadFrom:
    def get_some_sigs(self, location, *args, **kwargs):
        ss2 = utils.get_test_data("2.fa.sig")
        ss47 = utils.get_test_data("47.fa.sig")
        ss63 = utils.get_test_data("63.fa.sig")

        sig2 = sourmash.load_one_signature(ss2, ksize=31)
        sig47 = sourmash.load_one_signature(ss47, ksize=31)
        sig63 = sourmash.load_one_signature(ss63, ksize=31)

        lidx = LinearIndex([sig2, sig47, sig63], location)

        return lidx

    get_some_sigs.priority = 1

    def setup_method(self):
        self.saved_plugins = plugins._plugin_load_from
        plugins._plugin_load_from = [
            FakeEntryPoint("test_load", self.get_some_sigs),
            FakeEntryPoint(
                "test_load", self.get_some_sigs, error_on_import=ModuleNotFoundError
            ),
        ]

    def teardown_method(self):
        plugins._plugin_load_from = self.saved_plugins

    def test_load_1(self):
        ps = list(plugins.get_load_from_functions())
        assert len(ps) == 1

    def test_load_2(self, runtmp):
        fake_location = runtmp.output("passed-through location")
        idx = sourmash.load_file_as_index(fake_location)
        print(idx, idx.location)

        assert len(idx) == 3
        assert idx.location == fake_location


class Test_EntryPoint_LoadFrom_Priority:
    def get_some_sigs(self, location, *args, **kwargs):
        ss2 = utils.get_test_data("2.fa.sig")
        ss47 = utils.get_test_data("47.fa.sig")
        ss63 = utils.get_test_data("63.fa.sig")

        sig2 = sourmash.load_one_signature(ss2, ksize=31)
        sig47 = sourmash.load_one_signature(ss47, ksize=31)
        sig63 = sourmash.load_one_signature(ss63, ksize=31)

        lidx = LinearIndex([sig2, sig47, sig63], location)

        return lidx

    get_some_sigs.priority = 5

    def set_called_flag_1(self, location, *args, **kwargs):
        # high priority 1, raise ValueError
        print("setting flag 1")
        self.was_called_flag_1 = True
        raise ValueError

    set_called_flag_1.priority = 1

    def set_called_flag_2(self, location, *args, **kwargs):
        # high priority 2, return None
        print("setting flag 2")
        self.was_called_flag_2 = True

        return None

    set_called_flag_2.priority = 2

    def set_called_flag_3(self, location, *args, **kwargs):
        # lower priority 10, should not be called
        print("setting flag 3")
        self.was_called_flag_3 = True

        return None

    set_called_flag_3.priority = 10

    def setup_method(self):
        self.saved_plugins = plugins._plugin_load_from
        plugins._plugin_load_from = [
            FakeEntryPoint("test_load", self.get_some_sigs),
            FakeEntryPoint("test_load_2", self.set_called_flag_1),
            FakeEntryPoint("test_load_3", self.set_called_flag_2),
            FakeEntryPoint("test_load_4", self.set_called_flag_3),
        ]
        self.was_called_flag_1 = False
        self.was_called_flag_2 = False
        self.was_called_flag_3 = False

    def teardown_method(self):
        plugins._plugin_load_from = self.saved_plugins

    def test_load_1(self):
        ps = list(plugins.get_load_from_functions())
        assert len(ps) == 4

        assert not self.was_called_flag_1
        assert not self.was_called_flag_2
        assert not self.was_called_flag_3

    def test_load_2(self, runtmp):
        fake_location = runtmp.output("passed-through location")
        idx = sourmash.load_file_as_index(fake_location)
        print(idx, idx.location)

        assert len(idx) == 3
        assert idx.location == fake_location

        assert self.was_called_flag_1
        assert self.was_called_flag_2
        assert not self.was_called_flag_3


#
# Test basic features of the save_to plugin hook.
#


class FakeSaveClass(Base_SaveSignaturesToLocation):
    """
    A fake save class that just records what was sent to it.
    """

    priority = 50

    def __init__(self, location):
        super().__init__(location)
        self.keep = []

    @classmethod
    def matches(cls, location):
        if location:
            return location.endswith(".this-is-a-test")

    def add(self, ss):
        super().add(ss)
        self.keep.append(ss)


class FakeSaveClass_HighPriority(FakeSaveClass):
    priority = 1


class Test_EntryPointBasics_SaveTo:
    # test the basics
    def setup_method(self):
        self.saved_plugins = plugins._plugin_save_to
        plugins._plugin_save_to = [
            FakeEntryPoint("test_save", FakeSaveClass),
            FakeEntryPoint(
                "test_save", FakeSaveClass, error_on_import=ModuleNotFoundError
            ),
        ]

    def teardown_method(self):
        plugins._plugin_save_to = self.saved_plugins

    def test_save_1(self):
        ps = list(plugins.get_save_to_functions())
        print(ps)
        assert len(ps) == 1

    def test_save_2(self, runtmp):
        # load some signatures to save
        ss2 = utils.get_test_data("2.fa.sig")
        ss47 = utils.get_test_data("47.fa.sig")
        ss63 = utils.get_test_data("63.fa.sig")

        sig2 = sourmash.load_one_signature(ss2, ksize=31)
        sig47 = sourmash.load_one_signature(ss47, ksize=31)
        sig63 = sourmash.load_one_signature(ss63, ksize=31)

        # build a fake location that matches the FakeSaveClass
        # extension
        fake_location = runtmp.output("out.this-is-a-test")

        # this should use the plugin architecture to return an object
        # of type FakeSaveClass, with the three signatures in it.
        x = SaveSignaturesToLocation(fake_location)
        with x as save_sig:
            save_sig.add(sig2)
            save_sig.add(sig47)
            save_sig.add(sig63)

        print(len(x))
        print(x.keep)

        assert isinstance(x, FakeSaveClass)
        assert x.keep == [sig2, sig47, sig63]


class Test_EntryPointPriority_SaveTo:
    # test that priority is observed

    def setup_method(self):
        self.saved_plugins = plugins._plugin_save_to
        plugins._plugin_save_to = [
            FakeEntryPoint("test_save", FakeSaveClass),
            FakeEntryPoint("test_save2", FakeSaveClass_HighPriority),
        ]

    def teardown_method(self):
        plugins._plugin_save_to = self.saved_plugins

    def test_save_1(self):
        ps = list(plugins.get_save_to_functions())
        print(ps)
        assert len(ps) == 2

    def test_save_2(self, runtmp):
        # load some signatures to save
        ss2 = utils.get_test_data("2.fa.sig")
        ss47 = utils.get_test_data("47.fa.sig")
        ss63 = utils.get_test_data("63.fa.sig")

        sig2 = sourmash.load_one_signature(ss2, ksize=31)
        sig47 = sourmash.load_one_signature(ss47, ksize=31)
        sig63 = sourmash.load_one_signature(ss63, ksize=31)

        # build a fake location that matches the FakeSaveClass
        # extension
        fake_location = runtmp.output("out.this-is-a-test")

        # this should use the plugin architecture to return an object
        # of type FakeSaveClass, with the three signatures in it.
        x = SaveSignaturesToLocation(fake_location)
        with x as save_sig:
            save_sig.add(sig2)
            save_sig.add(sig47)
            save_sig.add(sig63)

        print(len(x))
        print(x.keep)

        assert isinstance(x, FakeSaveClass_HighPriority)
        assert x.keep == [sig2, sig47, sig63]
        assert x.priority == 1


#
# Test basic features of the save_to plugin hook.
#


class FakeCommandClass(plugins.CommandLinePlugin):
    """
    A fake CLI class.
    """

    command = "nifty"
    description = "do somethin' nifty"

    def __init__(self, parser):
        super().__init__(parser)
        parser.add_argument("arg1")
        parser.add_argument("--other", action="store_true")
        parser.add_argument("--do-fail", action="store_true")

    def main(self, args):
        super().main(args)
        print(f"hello, world! argument is: {args.arg1}")
        print(f"other is {args.other}")

        if args.do_fail:
            return 1
        return 0


class Test_EntryPointBasics_Command:
    # test the basics
    def setup_method(self):
        _ = plugins.get_cli_script_plugins()
        self.saved_plugins = plugins._plugin_cli
        plugins._plugin_cli_once = False
        plugins._plugin_cli = [FakeEntryPoint("test_command", FakeCommandClass)]

    def teardown_method(self):
        plugins._plugin_cli = self.saved_plugins

    def test_empty(self, runtmp):
        # empty out script plugins...
        plugins._plugin_cli = []

        with pytest.raises(utils.SourmashCommandFailed):
            runtmp.sourmash("scripts")
        out = runtmp.last_result.out
        err = runtmp.last_result.err
        print(out)
        print(err)
        assert "(No script plugins detected!)" in out

    def test_cmd_0(self, runtmp):
        # test default output with some plugins
        with pytest.raises(utils.SourmashCommandFailed):
            runtmp.sourmash("scripts")

        out = runtmp.last_result.out
        err = runtmp.last_result.err
        print(out)
        print(err)
        assert "do somethin' nifty" in out
        assert "sourmash scripts nifty" in out

    def test_cmd_1(self):
        # test descriptions
        ps = list(plugins.get_cli_scripts_descriptions())
        print(ps)
        assert len(ps) == 1

        descr0 = ps[0]
        assert "do somethin' nifty" in descr0
        assert "sourmash scripts nifty" in descr0

    def test_cmd_2(self):
        # test get_cli_script_plugins function
        ps = list(plugins.get_cli_script_plugins())
        print(ps)
        assert len(ps) == 1

    def test_cmd_3(self, runtmp):
        # test ability to run 'nifty' ;)
        with pytest.raises(utils.SourmashCommandFailed):
            runtmp.sourmash("scripts", "nifty")

        out = runtmp.last_result.out
        err = runtmp.last_result.err
        print(out)
        print(err)

        assert "nifty: error: the following arguments are required: arg1" in err
        assert "usage:  nifty [-h] [-q] [-d] [--other] [--do-fail] arg1" in err

    def test_cmd_4(self, runtmp):
        # test basic argument parsing etc
        runtmp.sourmash("scripts", "nifty", "--other", "some arg")

        out = runtmp.last_result.out
        err = runtmp.last_result.err
        print(out)
        print(err)

        assert "other is True" in out
        assert "hello, world! argument is: some arg" in out

    def test_cmd_5(self, runtmp):
        # test exit code passthru
        with pytest.raises(utils.SourmashCommandFailed):
            runtmp.sourmash("scripts", "nifty", "--do-fail", "some arg")

        status = runtmp.last_result.status
        out = runtmp.last_result.out
        err = runtmp.last_result.err
        print(out)
        print(err)
        print(status)

        assert "other is False" in out
        assert "hello, world! argument is: some arg" in out


class FakeCommandClass_Second(plugins.CommandLinePlugin):
    """
    A fake CLI class.
    """

    command = "more_nifty"
    description = "do somethin' else nifty"

    def __init__(self, parser):
        super().__init__(parser)
        parser.add_argument("arg1")
        parser.add_argument("--other", action="store_true")
        parser.add_argument("--do-fail", action="store_true")

    def main(self, args):
        super().main(args)
        print(f"hello, world! argument is: {args.arg1}")
        print(f"other is {args.other}")

        if args.do_fail:
            return 1
        return 0


class FakeCommandClass_Broken_1:
    """
    A fake CLI class.
    """

    # command = 'more_nifty' # no command

    def __init__(self, parser):
        assert 0

    def main(self, args):
        assert 0


class FakeCommandClass_Broken_2:
    """
    A fake CLI class.
    """

    command = "broken"
    # no description

    def __init__(self, parser):
        pass

    def main(self, args):
        return 0


class Test_EntryPointBasics_TwoCommands:
    # test a second command
    def setup_method(self):
        _ = plugins.get_cli_script_plugins()
        self.saved_plugins = plugins._plugin_cli
        plugins._plugin_cli_once = False
        plugins._plugin_cli = [
            FakeEntryPoint("test_command", FakeCommandClass),
            FakeEntryPoint("test_command2", FakeCommandClass_Second),
            FakeEntryPoint("test_command3", FakeCommandClass_Broken_1),
            FakeEntryPoint("test_command4", FakeCommandClass_Broken_2),
            FakeEntryPoint(
                "error-on-import", FakeCommandClass, error_on_import=ModuleNotFoundError
            ),
        ]

    def teardown_method(self):
        plugins._plugin_cli = self.saved_plugins

    def test_cmd_0(self, runtmp):
        # test default output for a few plugins
        with pytest.raises(utils.SourmashCommandFailed):
            runtmp.sourmash("scripts")

        out = runtmp.last_result.out
        err = runtmp.last_result.err
        print(out)
        print(err)
        assert "do somethin' nifty" in out
        assert "sourmash scripts nifty" in out

        assert "do somethin' else nifty" in out
        assert "sourmash scripts more_nifty" in out

    def test_cmd_1(self, runtmp):
        # test 'nifty'
        runtmp.sourmash("scripts", "nifty", "some arg")

        status = runtmp.last_result.status
        out = runtmp.last_result.out
        err = runtmp.last_result.err
        print(out)
        print(err)
        print(status)

        assert "other is False" in out
        assert "hello, world! argument is: some arg" in out

    def test_cmd_2(self, runtmp):
        # test 'more_nifty'
        runtmp.sourmash("scripts", "more_nifty", "some arg")

        status = runtmp.last_result.status
        out = runtmp.last_result.out
        err = runtmp.last_result.err
        print(out)
        print(err)
        print(status)

        assert "other is False" in out
        assert "hello, world! argument is: some arg" in out

    def test_sourmash_info(self, runtmp):
        # test 'sourmash info -v' => shows the plugins
        runtmp.sourmash("info", "-v")

        out = runtmp.last_result.out
        err = runtmp.last_result.err
        print(out)
        print(err)

        expected = """
groupfoo             test_plugin_framework          0.1   test_command
groupfoo             test_plugin_framework          0.1   test_command2
groupfoo             test_plugin_framework          0.1   test_command3
groupfoo             test_plugin_framework          0.1   test_command4
""".splitlines()
        for line in expected:
            assert line in err


def test_cli_scripts_getattr_fail():
    # test scripts.__getattr__ w/fail
    from sourmash.cli import scripts

    with pytest.raises(AttributeError):
        scripts.ThisAttrDoesNotExist


def test_cli_scripts_getattr_succ():
    # test scripts.__getattr__ w/success
    from sourmash.cli import scripts

    scripts.subparser



# This is a simple function designed to raise a warning whenever the sketch size is too small
# for the set being estimated. See the discussion in issue #1798
import numpy as np


def set_size_chernoff(set_size, scale, relative_error=0.05):
    """
    Computes the probability that the estimate: sketch_size * scaled deviates from the true
    set_size by more than relative_error. This relies on the fact that the sketch_size
    is binomially distributed with parameters sketch_size and 1/scale. The two-sided Chernoff
    bounds are used.
    @param set_size: The number of distinct k-mers in the given set
    @param relative_error: the desired relative error (defaults to 5%)
    @return: float (the upper bound probability)
    """
    upper_bound = 1 - 2 * np.exp(-(relative_error**2) * set_size / (scale * 3))
    return upper_bound


def get_set_size(scale, num_sketches):
    """
    This returns the expected number of distinct k-mers from the scale size and number of sketches.
    @param scale: the scale factor used
    @param num_sketches: the actual size of the FracMinHash sketch
    @return: int (the expected number of distinct k-mers in the original set of k-mers)
    """
    #  TODO: replace with HLL when that gets implemented
    return int(np.floor(scale * num_sketches))


def set_size_estimate_is_accurate(
    scale, num_sketches, relative_error=0.05, confidence=0.95
):
    set_size = get_set_size(scale, num_sketches)
    probability = set_size_chernoff(set_size, scale, relative_error)
    if probability >= confidence:
        return True
    else:
        return False


def test_set_size_chernoff():
    eps = 10 ** (-6)
    rel_error = 0.01
    set_size = 1000000
    s = 1 / 0.1  # I'm used to using a scale value between 0 and 1
    value_from_mathematica = 0.928652
    assert (
        np.abs(set_size_chernoff(set_size, s, rel_error) - value_from_mathematica) < eps
    )

    rel_error = 0.05
    set_size = 10000
    s = 1
    value_from_mathematica = 0.999519
    assert (
        np.abs(set_size_chernoff(set_size, s, rel_error) - value_from_mathematica) < eps
    )

    rel_error = 0.001
    set_size = 10
    s = 1 / 0.01
    value_from_mathematica = -1
    assert (
        np.abs(set_size_chernoff(set_size, s, rel_error) - value_from_mathematica) < eps
    )


def test_set_size_estimate_is_accurate():
    10 ** (-6)
    rel_error = 0.05
    set_size = 1000000
    s = 1 / 0.1  # I'm used to using a scale value between 0 and 1
    num_sketches = set_size / s  # idealized case
    confidence = 0.95
    assert (
        set_size_estimate_is_accurate(
            scale=s,
            num_sketches=num_sketches,
            relative_error=rel_error,
            confidence=confidence,
        )
        is True
    )
    confidence = set_size_chernoff(set_size=set_size, scale=s, relative_error=rel_error)
    assert (
        set_size_estimate_is_accurate(
            scale=s,
            num_sketches=num_sketches,
            relative_error=rel_error,
            confidence=confidence,
        )
        is True
    )
    # Horrible values
    assert (
        set_size_estimate_is_accurate(
            scale=10000, num_sketches=num_sketches, relative_error=0, confidence=1
        )
        is False
    )
    # Less horrible, but still bad values
    confidence = set_size_chernoff(set_size=set_size, scale=s, relative_error=rel_error)
    assert (
        set_size_estimate_is_accurate(
            scale=s,
            num_sketches=num_sketches,
            relative_error=rel_error,
            confidence=confidence * 2,
        )
        is False
    )
    # one where the confidence is negative
    rel_error = 0.001
    set_size = 10
    s = 100
    num_sketches = set_size / s
    assert (
        set_size_estimate_is_accurate(
            scale=s,
            num_sketches=num_sketches,
            relative_error=rel_error,
            confidence=confidence,
        )
        is False
    )
    assert (
        set_size_estimate_is_accurate(
            scale=s, num_sketches=0, relative_error=rel_error, confidence=confidence
        )
        is False
    )


def run_tests():
    test_set_size_chernoff()
    test_set_size_estimate_is_accurate()


if __name__ == "__main__":
    print("Running tests")
    run_tests()
    print("Tests completed successfully")
